from typing import List

import math
import torch
from torch import nn


class MLP(nn.Module):
    """A simple MLP.

    Flattens all dimensions except batch and uses GELU nonlinearities.
    """

    def __init__(
        self, in_shape, out_shape, hidden_channels, hidden_layers, dropout_prob=None
    ):
        super().__init__()

        if not hidden_layers > 0:
            raise NotImplementedError("Only supports > 0 hidden layers")

        self.in_shape = in_shape
        self.out_shape = out_shape

        layers: List[nn.Module] = [nn.Linear(prod(in_shape), hidden_channels)]
        if dropout_prob is not None:
            layers.append(nn.Dropout(dropout_prob))
        for _ in range(hidden_layers - 1):
            layers.append(nn.GELU())
            layers.append(nn.Linear(hidden_channels, hidden_channels))
            if dropout_prob is not None:
                layers.append(nn.Dropout(dropout_prob))

        layers.append(nn.GELU())
        layers.append(nn.Linear(hidden_channels, prod(self.out_shape)))
        self.mlp = nn.Sequential(*layers)

    def forward(self, inputs: torch.Tensor):
        """Forward pass of MLP."""
        return self.mlp(inputs)


def prod(shape):
    if isinstance(shape, int):
        return shape
    else:
        return math.prod(shape)



import torch
import pytest
from tests.constants import (
    TOLERANCES,
    MILD_TOLERANCES,
    LOGM2_MEAN_STD,
    REPS,
    FRAMES_PREDICTOR,
)
from tests.helpers import sample_particle, equivectors_builder
from tests.hep import get_tagging_features

from lloca.backbone.particletransformer import ParticleTransformer, Block
from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.utils.rand_transforms import rand_lorentz
from lloca.framesnet.frames import InverseFrames
from lloca.framesnet.equi_frames import LearnedSO13Frames, LearnedPDFrames
from lloca.backbone.attention import LLoCaAttention


@pytest.mark.parametrize("FramesPredictor", FRAMES_PREDICTOR)
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("num_heads", [8])
@pytest.mark.parametrize("attn_reps", REPS)
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_block_invariance_equivariance(
    FramesPredictor,
    batch_dims,
    logm2_std,
    logm2_mean,
    attn_reps,
    num_heads,
):
    dtype = torch.float64

    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # define block
    in_reps = TensorReps("1x1n")
    attn_reps = TensorReps(attn_reps)
    trafo = TensorRepsTransform(TensorReps(in_reps))
    linear_in = torch.nn.Linear(in_reps.dim, attn_reps.dim * num_heads).to(dtype=dtype)
    linear_out = torch.nn.Linear(attn_reps.dim * num_heads, in_reps.dim).to(dtype=dtype)
    attention = LLoCaAttention(attn_reps, num_heads)
    ParT_block = Block(attention=attention, embed_dim=attn_reps.dim * num_heads).to(
        dtype
    )
    ParT_block.eval()  # turn off dropout

    def block_wrapper(x, frames):
        x = x.unsqueeze(0)
        mask = torch.ones_like(x[..., 0])
        frames = frames.reshape(1, *frames.shape)
        attention.prepare_frames(frames)
        x = ParT_block(x=x, padding_mask=mask)
        x = x.squeeze(0)
        return x

    # get global transformation
    random = rand_lorentz([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    frames = call_predictor(fm)
    fm_local = trafo(fm, frames)

    # block - global
    x_local = linear_in(fm_local)
    x_prime_local = block_wrapper(x_local, frames)
    fm_prime_local = linear_out(x_prime_local)
    # back to global
    fm_prime_global = trafo(fm_prime_local, InverseFrames(frames))
    fm_prime_tr_global = torch.einsum("...ij,...j->...i", random, fm_prime_global)

    # global - block
    fm_transformed = torch.einsum("...ij,...j->...i", random, fm)
    frames_transformed = call_predictor(fm_transformed)
    fm_tr_local = trafo(fm_transformed, frames_transformed)
    x_tr_local = linear_in(fm_tr_local)
    x_tr_prime_local = block_wrapper(x_tr_local, frames_transformed)
    fm_tr_prime_local = linear_out(x_tr_prime_local)
    # back to global frame
    fm_tr_prime_global = trafo(fm_tr_prime_local, InverseFrames(frames_transformed))

    # test feature invariance before the operation
    torch.testing.assert_close(x_local, x_tr_local, **TOLERANCES)

    # test feature invariance after the operation
    torch.testing.assert_close(x_tr_prime_local, x_prime_local, **TOLERANCES)

    # test equivariance of outputs
    torch.testing.assert_close(fm_tr_prime_global, fm_prime_tr_global, **TOLERANCES)


@pytest.mark.parametrize(
    "FramesPredictor", [LearnedSO13Frames, LearnedPDFrames]
)  # RestFrames gives nans sometimes
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_ParT_invariance(
    FramesPredictor,
    batch_dims,
    logm2_std,
    logm2_mean,
):
    dtype = torch.float64
    batch = torch.zeros(batch_dims[0], dtype=torch.long)

    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # define ParT
    in_reps = TensorReps("1x1n")
    trafo = TensorRepsTransform(TensorReps(in_reps))
    model = ParticleTransformer(
        input_dim=7,
        num_classes=1,
        attn_reps="8x0n+2x1n",
    ).to(dtype=dtype)
    model.eval()  # turn off dropout

    def ParT_wrapper(p_local, frames):
        fts_local = get_tagging_features(p_local, batch)
        fts_local = fts_local.transpose(-1, -2).unsqueeze(0)
        p_local = p_local[..., [1, 2, 3, 0]]
        p_local = p_local.transpose(-1, -2).unsqueeze(0)
        mask = torch.ones_like(p_local[..., [0], :])
        frames = frames.reshape(1, *frames.shape)
        x = model(x=fts_local, v=p_local, frames=frames, mask=mask)
        x = x.transpose(-1, -2).squeeze(0)
        return x

    # get global transformation
    random = rand_lorentz([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    frames = call_predictor(fm)
    fm_local = trafo(fm, frames)

    # ParT
    score_prime_local = ParT_wrapper(fm_local, frames)

    # global - ParT
    fm_transformed = torch.einsum("...ij,...j->...i", random, fm)
    frames_transformed = call_predictor(fm_transformed)
    fm_tr_local = trafo(fm_transformed, frames_transformed)
    score_tr_prime_local = ParT_wrapper(fm_tr_local, frames_transformed)

    # test feature invariance before the operation
    torch.testing.assert_close(fm_local, fm_tr_local, **TOLERANCES)

    # test equivariance of scores
    torch.testing.assert_close(
        score_tr_prime_local, score_prime_local, **MILD_TOLERANCES
    )



import torch


def lorentz_inner(v1, v2):
    """Lorentz inner product, i.e v1*g*v2

    Parameters
    ----------
    v1, v2 : torch.Tensor
        Tensors of shape (*dims, 4)

    Returns
    -------
    torch.Tensor
        Lorentz inner product of shape (*dims, )
    """
    prod = v1 * v2
    prod *= torch.tensor([1, -1, -1, -1], device=v1.device, dtype=v1.dtype)
    return prod.sum(dim=-1)


def lorentz_squarednorm(v):
    """Lorentz norm, i.e. v*g*v

    Parameters
    ----------
    v : torch.Tensor
        Tensor of shape (*dims, 4)

    Returns
    -------
    torch.Tensor
        Lorentz norm of shape (*dims, )
    """
    return lorentz_inner(v, v)


def lorentz_eye(dims, device=torch.device("cpu"), dtype=torch.float32):
    """
    Create a identity matrix of given shape

    Parameters
    ----------
    dims : tuple
        Dimensions of the output tensor, e.g. (2, 3) for a 2x3 matrix
    device : torch.device, optional
        Device to create the tensor on, by default torch.device("cpu")
    dtype : torch.dtype, optional
        Data type of the tensor, by default torch.float32

    Returns
    -------
    torch.Tensor
        Identity matrix of shape (*dims, 4, 4)
    """
    eye = torch.eye(4, dtype=dtype, device=device)
    eye = eye.view((1,) * len(dims) + eye.shape).repeat(*dims, 1, 1)
    return eye


def lorentz_metric(dims, device=torch.device("cpu"), dtype=torch.float32):
    """
    Create a metric tensor of given shape

    Parameters
    ----------
    dims : tuple
        Dimensions of the output tensor, e.g. (2, 3) for a 2x3 matrix
    device : torch.device, optional
        Device to create the tensor on, by default torch.device("cpu")
    dtype : torch.dtype, optional
        Data type of the tensor, by default torch.float32

    Returns
    -------
    torch.Tensor
        Metric tensor of shape (*dims, 4, 4)
    """
    eye = torch.eye(4, device=device, dtype=dtype)
    eye[1:, 1:] *= -1
    eye = eye.view((1,) * len(dims) + eye.shape).repeat(*dims, 1, 1)
    return eye


def lorentz_cross(v1, v2, v3):
    """
    Compute the cross product in Minkowski space.

    Parameters
    ----------
    v1, v2, v3 : torch.Tensor
        Tensors of shape (*dims, 4) representing vectors in Minkowski space.

    Returns
    -------
    torch.Tensor
        The cross product of the three vectors, shape (*dims, 4).
    """
    assert v1.shape[-1] == 4
    assert v1.shape == v2.shape and v1.shape == v3.shape

    mat = torch.stack([v1, v2, v3], dim=-1)

    # euclidean fully antisymmetric product
    v4 = []
    for n in range(4):
        minor = torch.cat([mat[..., :n, :], mat[..., n + 1 :, :]], dim=-2)
        contribution = (-1) ** n * torch.det(minor)
        v4.append(contribution)
    v4 = torch.stack(v4, dim=-1)

    # raise indices with metric tensor
    v4 *= torch.tensor([1.0, -1.0, -1.0, -1.0], device=v1.device, dtype=v1.dtype)
    return v4



import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from .lloca_message_passing import LLoCaMessagePassing
from .mlp import MLP
from ..reps.tensorreps import TensorReps


class EdgeConv(LLoCaMessagePassing):
    def __init__(
        self,
        reps,
        num_layers_mlp1,
        num_layers_mlp2,
        aggr="add",
        num_edge_attr=0,
        dropout_prob=None,
    ):
        super().__init__(aggr=aggr, params_dict={"x": {"type": "local", "rep": reps}})
        self.mlp1 = MLP(
            in_shape=[reps.dim * 2 + num_edge_attr],
            out_shape=[reps.dim],
            hidden_layers=num_layers_mlp1,
            hidden_channels=reps.dim,
            dropout_prob=dropout_prob,
        )
        self.mlp2 = (
            MLP(
                in_shape=[reps.dim],
                out_shape=[reps.dim],
                hidden_layers=num_layers_mlp2,
                hidden_channels=reps.dim,
                dropout_prob=dropout_prob,
            )
            if num_layers_mlp2 > 0
            else nn.Identity()
        )

    def forward(self, x, frames, edge_index, batch=None, edge_attr=None):
        frames = (frames, frames)

        x_aggr = self.propagate(
            edge_index,
            x=x,
            frames=frames,
            edge_attr=edge_attr,
            batch=batch,
        )
        x_aggr = self.mlp2(x_aggr)
        return x_aggr

    def message(self, x_i, x_j, frames_i, frames_j, edge_attr=None):
        x = x_j
        x = torch.cat((x, x_i), dim=-1)
        if edge_attr is not None:
            x = torch.cat((x, edge_attr), dim=-1)
        x = self.mlp1(x)
        return x


class GraphNet(nn.Module):
    """Baseline graphnet.

    Combines num_blocks EdgeConv blocks.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    hidden_reps : str
        Representation during message passing.
    out_channels : int
        Number of output channels.
    num_blocks : int
        Number of EdgeConv blocks.
    """

    def __init__(
        self,
        in_channels: int,
        hidden_reps: str,
        out_channels: int,
        num_blocks: int,
        *args,
        checkpoint_blocks=False,
        **kwargs,
    ):
        super().__init__()
        hidden_reps = TensorReps(hidden_reps)
        self.checkpoint_blocks = checkpoint_blocks

        self.linear_in = nn.Linear(in_channels, hidden_reps.dim)
        self.linear_out = nn.Linear(hidden_reps.dim, out_channels)
        self.blocks = nn.ModuleList(
            [
                EdgeConv(
                    hidden_reps,
                    *args,
                    **kwargs,
                )
                for _ in range(num_blocks)
            ]
        )

    def forward(self, inputs, frames, edge_index, batch=None, edge_attr=None):
        """Forward pass.

        Parameters
        ----------
        inputs : Tensor
            Input data with shape (num_items, in_channels)
        frames : Frames
            Local frames used for message passing
        edge_index : Tensor
            Edge index tensor with shape (2, num_edges)
        batch : Tensor
            Batch tensorwith shape (num_items,)
            If None, assumes fully connected graph along the num_items direction.
        edge_attr : Tensor
            Edge attribute tensor with shape (num_items, num_edge_attr)

        Returns
        -------
        outputs : Tensor
            Outputs with shape (num_items, out_channels)
        """
        x = self.linear_in(inputs)
        for block in self.blocks:
            if self.checkpoint_blocks:
                x = checkpoint(
                    block,
                    x=x,
                    frames=frames,
                    edge_index=edge_index,
                    batch=batch,
                    edge_attr=edge_attr,
                )
            else:
                x = block(
                    x=x,
                    frames=frames,
                    edge_index=edge_index,
                    batch=batch,
                    edge_attr=edge_attr,
                )
        outputs = self.linear_out(x)
        return outputs



"""ParticleNet

Paper: "ParticleNet: Jet Tagging via Particle Clouds" - https://arxiv.org/abs/1902.08570
Code: https://github.com/hqucms/weaver-core/blob/main/weaver/nn/model/ParticleNet.py

We have to do three things to build a LLoCa-ParticleNet
- Specify the hidden_reps_list for hidden representations in each message-passing layer
- Pass the frames through the network
- During the edge convolution, transform the features from the local to the global frames
  using tools from the lloca code (TensorRepsTransform, IndexSelectFrames, ChangeOfFrames)

Use 'git diff --no-index experiments/baselines/particlenet.py lloca/nn/particlenet.py'
to see the changes required to include frame-to-frame transformations.
"""

import torch
import torch.nn as nn

from ..reps.tensorreps import TensorReps
from ..reps.tensorreps_transform import TensorRepsTransform
from ..framesnet.frames import IndexSelectFrames, ChangeOfFrames


def change_local_frame(x_j_framej, idx, frames, trafo):
    """Transform features x_j from frame 'j' ('x_j_framej') to frame 'i' ('x_j_framei').

    Parameters
    ----------
    x_j_framej : torch.Tensor
        Input features in local frame 'j' of shape (batch_size, num_dims, num_points, k).
    idx : torch.Tensor
        Indices of the nearest neighbors in the batch of shape (batch_size*num_points*k).
    frames : Frames
        Local frames of reference for the particles, shape (num_points, 4, 4).
    trafo : TensorRepsTransform
        Transformation function to apply to the features.

    Returns
    -------
    torch.Tensor
    """
    # we use batch_size*num_points with repeats of k for idx_i, e.g. for 2 points with 3 batch and k=2,
    # idx_i becomes (0,1,2,3,4,5) -> (0,0,1,1,2,2,3,3,4,4,5,5).
    idx_i = torch.arange(
        x_j_framej.shape[2] * x_j_framej.shape[0], device=x_j_framej.device
    ).repeat_interleave(
        x_j_framej.shape[-1]
    )  # identity (batch, num_points*k)
    idx_j = idx  # indices from knn (batch, num_points*k)

    frames_i = IndexSelectFrames(frames, idx_i)
    frames_j = IndexSelectFrames(frames, idx_j)
    trafo_j_to_i = ChangeOfFrames(
        frames_j, frames_i
    )  # convention: (frames_start, frames_end)

    # reshape and apply trafo
    x_j_framej_2 = x_j_framej.permute(
        0, 2, 3, 1
    )  # (batch_size, num_points, k, num_dims)
    pre = x_j_framej_2.reshape(
        -1, x_j_framej_2.shape[-1]
    )  # (batch_size*num_points*k, num_dims)
    x_j_framei = trafo(pre, trafo_j_to_i)
    x_j_framei = x_j_framei.view(x_j_framej_2.shape).permute(
        0, 3, 1, 2
    )  # (batch_size, num_dims, num_points, k)
    return x_j_framei


def knn(x, k):
    inner = -2 * torch.matmul(x.transpose(2, 1), x)
    xx = torch.sum(x**2, dim=1, keepdim=True)
    pairwise_distance = -xx - inner - xx.transpose(2, 1)
    idx = pairwise_distance.topk(k=k + 1, dim=-1)[1][
        :, :, 1:
    ]  # (batch_size, num_points, k)
    return idx


# v1 is faster on GPU
def get_graph_feature_v1(x, k, idx, frames, trafo):
    batch_size, num_dims, num_points = x.size()

    idx_base = torch.arange(0, batch_size, device=x.device).view(-1, 1, 1) * num_points
    idx = idx + idx_base
    idx = idx.view(-1)

    fts = x.transpose(2, 1).reshape(
        -1, num_dims
    )  # -> (batch_size, num_points, num_dims) -> (batch_size*num_points, num_dims)
    fts = fts[idx, :].view(
        batch_size, num_points, k, num_dims
    )  # neighbors: -> (batch_size*num_points*k, num_dims) -> ...
    fts = fts.permute(0, 3, 1, 2).contiguous()  # (batch_size, num_dims, num_points, k)
    x = x.view(batch_size, num_dims, num_points, 1).repeat(1, 1, 1, k)
    fts = change_local_frame(fts, idx, frames, trafo)
    fts = torch.cat((x, fts - x), dim=1)  # ->(batch_size, 2*num_dims, num_points, k)
    return fts


# v2 is faster on CPU
def get_graph_feature_v2(x, k, idx, frames, trafo):
    batch_size, num_dims, num_points = x.size()

    idx_base = torch.arange(0, batch_size, device=x.device).view(-1, 1, 1) * num_points
    idx = idx + idx_base
    idx = idx.view(-1)

    fts = x.transpose(0, 1).reshape(
        num_dims, -1
    )  # -> (num_dims, batch_size, num_points) -> (num_dims, batch_size*num_points)
    fts = fts[:, idx].view(
        num_dims, batch_size, num_points, k
    )  # neighbors: -> (num_dims, batch_size*num_points*k) -> ...
    fts = fts.transpose(1, 0).contiguous()  # (batch_size, num_dims, num_points, k)
    fts = change_local_frame(fts, idx, frames, trafo)

    x = x.view(batch_size, num_dims, num_points, 1).repeat(1, 1, 1, k)
    fts = torch.cat((x, fts - x), dim=1)  # ->(batch_size, 2*num_dims, num_points, k)

    return fts


class EdgeConvBlock(nn.Module):
    r"""EdgeConv layer.
    Introduced in "`Dynamic Graph CNN for Learning on Point Clouds
    <https://arxiv.org/pdf/1801.07829>`__".  Can be described as follows:
    .. math::
       x_i^{(l+1)} = \max_{j \in \mathcal{N}(i)} \mathrm{ReLU}(
       \Theta \cdot (x_j^{(l)} - x_i^{(l)}) + \Phi \cdot x_i^{(l)})
    where :math:`\mathcal{N}(i)` is the neighbor of :math:`i`.
    Parameters
    ----------
    in_feat : int
        Input feature size.
    out_feat : int
        Output feature size.
    batch_norm : bool
        Whether to include batch normalization on messages.
    """

    def __init__(
        self,
        k,
        in_reps,
        out_feats,
        batch_norm=True,
        activation=True,
        cpu_mode=False,
    ):
        super(EdgeConvBlock, self).__init__()
        self.k = k
        self.batch_norm = batch_norm
        self.activation = activation
        self.num_layers = len(out_feats)
        self.get_graph_feature = (
            get_graph_feature_v2 if cpu_mode else get_graph_feature_v1
        )
        in_feat = in_reps.dim
        self.trafo = TensorRepsTransform(TensorReps(in_reps))

        self.convs = nn.ModuleList()
        for i in range(self.num_layers):
            self.convs.append(
                nn.Conv2d(
                    2 * in_feat if i == 0 else out_feats[i - 1],
                    out_feats[i],
                    kernel_size=1,
                    bias=False if self.batch_norm else True,
                )
            )

        if batch_norm:
            self.bns = nn.ModuleList()
            for i in range(self.num_layers):
                self.bns.append(nn.BatchNorm2d(out_feats[i]))

        if activation:
            self.acts = nn.ModuleList()
            for i in range(self.num_layers):
                self.acts.append(nn.ReLU())

        if in_feat == out_feats[-1]:
            self.sc = None
        else:
            self.sc = nn.Conv1d(in_feat, out_feats[-1], kernel_size=1, bias=False)
            self.sc_bn = nn.BatchNorm1d(out_feats[-1])

        if activation:
            self.sc_act = nn.ReLU()

    def forward(self, points, features, frames):

        topk_indices = knn(points, self.k)
        x = self.get_graph_feature(features, self.k, topk_indices, frames, self.trafo)

        for conv, bn, act in zip(self.convs, self.bns, self.acts):
            x = conv(x)  # (N, C', P, K)
            if bn:
                x = bn(x)
            if act:
                x = act(x)

        fts = x.mean(dim=-1)  # (N, C, P)

        # shortcut
        if self.sc:
            sc = self.sc(features)  # (N, C_out, P)
            sc = self.sc_bn(sc)
        else:
            sc = features

        return self.sc_act(sc + fts)  # (N, C_out, P)


class ParticleNet(nn.Module):
    def __init__(
        self,
        input_dims,
        hidden_reps_list,
        num_classes,
        conv_params=[(7, (32, 32, 32)), (7, (64, 64, 64))],
        fc_params=[(128, 0.1)],
        use_fusion=True,
        use_fts_bn=True,
        use_counts=True,
        for_inference=False,
        for_segmentation=False,
        **kwargs
    ):
        # hidden_reps_list: hidden representation for message-passing at beginning of each layer
        super(ParticleNet, self).__init__(**kwargs)
        hidden_reps_list = [TensorReps(x) for x in hidden_reps_list]
        assert input_dims == hidden_reps_list[0].dim
        assert len(hidden_reps_list) == len(conv_params)

        self.use_fts_bn = use_fts_bn
        if self.use_fts_bn:
            self.bn_fts = nn.BatchNorm1d(hidden_reps_list[0].dim)

        self.use_counts = use_counts

        self.edge_convs = nn.ModuleList()
        for idx, layer_param in enumerate(conv_params):
            k, channels = layer_param
            in_reps = hidden_reps_list[idx]
            assert (
                in_reps.dim == conv_params[idx - 1][1][-1]
                if idx > 0
                else hidden_reps_list[0].dim
            )
            self.edge_convs.append(
                EdgeConvBlock(
                    k=k, in_reps=in_reps, out_feats=channels, cpu_mode=for_inference
                )
            )

        self.use_fusion = use_fusion
        if self.use_fusion:
            in_chn = sum(x[-1] for _, x in conv_params)
            out_chn = max(128, min((in_chn // 128) * 128, 1024))
            self.fusion_block = nn.Sequential(
                nn.Conv1d(in_chn, out_chn, kernel_size=1, bias=False),
                nn.BatchNorm1d(out_chn),
                nn.ReLU(),
            )

        self.for_segmentation = for_segmentation

        fcs = []
        for idx, layer_param in enumerate(fc_params):
            channels, drop_rate = layer_param
            if idx == 0:
                in_chn = out_chn if self.use_fusion else conv_params[-1][1][-1]
            else:
                in_chn = fc_params[idx - 1][0]
            if self.for_segmentation:
                fcs.append(
                    nn.Sequential(
                        nn.Conv1d(in_chn, channels, kernel_size=1, bias=False),
                        nn.BatchNorm1d(channels),
                        nn.ReLU(),
                        nn.Dropout(drop_rate),
                    )
                )
            else:
                fcs.append(
                    nn.Sequential(
                        nn.Linear(in_chn, channels), nn.ReLU(), nn.Dropout(drop_rate)
                    )
                )
        if self.for_segmentation:
            fcs.append(nn.Conv1d(fc_params[-1][0], num_classes, kernel_size=1))
        else:
            fcs.append(nn.Linear(fc_params[-1][0], num_classes))
        self.fc = nn.Sequential(*fcs)

        self.for_inference = for_inference

    def forward(self, points, features, frames, mask=None):
        #         print('points:\n', points)
        #         print('features:\n', features)
        if mask is None:
            mask = features.abs().sum(dim=1, keepdim=True) != 0  # (N, 1, P)
        points *= mask
        features *= mask
        coord_shift = (mask == 0) * 1e9
        if self.use_counts:
            counts = mask.float().sum(dim=-1)
            counts = torch.max(counts, torch.ones_like(counts))  # >=1

        if self.use_fts_bn:
            fts = self.bn_fts(features) * mask
        else:
            fts = features
        outputs = []
        for idx, conv in enumerate(self.edge_convs):
            pts = (points if idx == 0 else fts) + coord_shift
            fts = conv(pts, fts, frames) * mask
            if self.use_fusion:
                outputs.append(fts)
        if self.use_fusion:
            fts = self.fusion_block(torch.cat(outputs, dim=1)) * mask

        #         assert(((fts.abs().sum(dim=1, keepdim=True) != 0).float() - mask.float()).abs().sum().item() == 0)

        if self.for_segmentation:
            x = fts
        else:
            if self.use_counts:
                x = fts.sum(dim=-1) / counts  # divide by the real counts
            else:
                x = fts.mean(dim=-1)

        output = self.fc(x)
        if self.for_inference:
            output = torch.softmax(output, dim=1)
        # print('output:\n', output)
        return output



import torch

EPS = 1e-10
CUTOFF = 10

# weaver defaults for tagging features standardization (mean, std)
TAGGING_FEATURES_PREPROCESSING = [
    [1.7, 0.7],  # log_pt
    [2.0, 0.7],  # log_energy
    [-4.7, 0.7],  # log_pt_rel
    [-4.7, 0.7],  # log_energy_rel
    [0, 1],  # dphi
    [0, 1],  # deta
    [0.2, 4],  # dr
]


def get_pt(p):
    # transverse momentum
    return torch.sqrt((p[..., 1] ** 2 + p[..., 2] ** 2).clamp(min=EPS))


def stable_arctanh(x, eps=1e-10):
    # implementation of arctanh that avoids log(0) issues
    return 0.5 * (torch.log((1 + x).clamp(min=eps)) - torch.log((1 - x).clamp(min=eps)))


def avoid_zero(x, eps=EPS):
    # set small-abs values to eps for numerical stability
    return torch.where(x.abs() < eps, eps, x)


def get_phi(p):
    # azimuthal angle
    return torch.arctan2(avoid_zero(p[..., 2]), avoid_zero(p[..., 1]))


def get_eta(p):
    # rapidity
    p_abs = torch.sqrt(torch.sum(p[..., 1:] ** 2, dim=-1).clamp(min=EPS))
    return stable_arctanh(p[..., 3] / p_abs)


def get_tagging_features(fourmomenta, jet, eps=1e-10):
    """
    Compute features typically used in jet tagging

    Parameters
    ----------
    fourmomenta: torch.tensor of shape (n_particles, 4)
        Fourmomenta in the format (E, px, py, pz)
    jet: torch.tensor of shape (n_particles, 4)
        Jet momenta in the shape (E, px, py, pz)
    eps: float

    Returns
    -------
    features: torch.tensor of shape (n_particles, 7)
        Features: log_pt, log_energy, log_pt_rel, log_energy_rel, dphi, deta, dr
    """
    log_pt = get_pt(fourmomenta).unsqueeze(-1).log()
    log_energy = fourmomenta[..., 0].unsqueeze(-1).clamp(min=eps).log()

    log_pt_rel = (get_pt(fourmomenta).log() - get_pt(jet).log()).unsqueeze(-1)
    log_energy_rel = (
        fourmomenta[..., 0].clamp(min=eps).log() - jet[..., 0].clamp(min=eps).log()
    ).unsqueeze(-1)
    phi_4, phi_jet = get_phi(fourmomenta), get_phi(jet)
    dphi = ((phi_4 - phi_jet + torch.pi) % (2 * torch.pi) - torch.pi).unsqueeze(-1)
    eta_4, eta_jet = get_eta(fourmomenta), get_eta(jet)
    deta = -(eta_4 - eta_jet).unsqueeze(-1)
    dr = torch.sqrt((dphi**2 + deta**2).clamp(min=eps))
    features = [
        log_pt,
        log_energy,
        log_pt_rel,
        log_energy_rel,
        dphi,
        deta,
        dr,
    ]
    for i, feature in enumerate(features):
        mean, factor = TAGGING_FEATURES_PREPROCESSING[i]
        features[i] = (feature - mean) * factor
    features = torch.cat(features, dim=-1)
    return features



import torch
import pytest
from tests.constants import TOLERANCES, LOGM2_MEAN_STD
from tests.helpers import sample_particle

from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.framesnet.nonequi_frames import (
    IdentityFrames,
    RandomFrames,
)


@pytest.mark.parametrize(
    "FramesPredictor,transform_type",
    [
        (IdentityFrames, None),
        (RandomFrames, "lorentz"),
        (RandomFrames, "rotation"),
        (RandomFrames, "xyrotation"),
    ],
)
@pytest.mark.parametrize("batch_dims", [[1000]])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_vectors(FramesPredictor, transform_type, batch_dims, logm2_mean, logm2_std):
    dtype = torch.float32

    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)

    # predict local frames
    predictor = (
        FramesPredictor(transform_type=transform_type)
        if FramesPredictor == RandomFrames
        else FramesPredictor()
    )
    frames = predictor(fm)

    # transform into local frames
    reps = TensorReps("1x1n")
    trafo = TensorRepsTransform(TensorReps(reps))
    fm_local = trafo(fm, frames)

    if FramesPredictor == IdentityFrames:
        # fourmomenta should not change
        torch.testing.assert_close(fm_local, fm, **TOLERANCES)
    elif type == "rotation":
        # energy and pz should not change
        torch.testing.assert_close(fm_local[..., [0]], fm[..., [0]], **TOLERANCES)
    elif type == "xyrotation":
        # energy and pz should not change
        torch.testing.assert_close(fm_local[..., [0, 3]], fm[..., [0, 3]], **TOLERANCES)



from functools import partial
from typing import Optional

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from .attention import LLoCaAttention
from ..reps.tensorreps import TensorReps


class BaselineLayerNorm(nn.Module):
    """Baseline layer norm over all dimensions except the first."""

    @staticmethod
    def forward(inputs: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        inputs : Tensor
            Input data

        Returns
        -------
        outputs : Tensor
            Normalized inputs.
        """
        return torch.nn.functional.layer_norm(
            inputs, normalized_shape=inputs.shape[-1:]
        )


class MultiHeadQKVLinear(nn.Module):
    """Compute queries, keys, and values via multi-head attention.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    hidden_channels : int
        Number of hidden channels = size of query, key, and value.
    num_heads : int
        Number of attention heads.
    """

    def __init__(self, in_channels, hidden_channels, num_heads):
        super().__init__()
        self.num_heads = num_heads
        self.linear = nn.Linear(in_channels, 3 * hidden_channels * num_heads)

    def forward(self, inputs):
        """Forward pass.

        Returns
        -------
        q : Tensor
            Queries
        k : Tensor
            Keys
        v : Tensor
            Values
        """
        qkv = self.linear(inputs)  # (..., num_items, 3 * hidden_channels * num_heads)

        *leading, items, last = qkv.shape
        hidden_channels = last // (3 * self.num_heads)
        qkv = qkv.view(*leading, items, 3, hidden_channels, self.num_heads)
        qkv = qkv.movedim(-3, 0).movedim(-1, len(leading) + 1)
        q, k, v = qkv.unbind(dim=0)  # 3x (..., num_heads, num_items, hidden_channels)
        return q, k, v


class MultiQueryQKVLinear(nn.Module):
    """Compute queries, keys, and values via multi-query attention.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    hidden_channels : int
        Number of hidden channels = size of query, key, and value.
    num_heads : int
        Number of attention heads.
    """

    def __init__(self, in_channels, hidden_channels, num_heads):
        super().__init__()
        self.num_heads = num_heads
        self.q_linear = nn.Linear(in_channels, hidden_channels * num_heads)
        self.k_linear = nn.Linear(in_channels, hidden_channels)
        self.v_linear = nn.Linear(in_channels, hidden_channels)

    def forward(self, inputs):
        """Forward pass.

        Parameters
        ----------
        inputs : Tensor
            Input data

        Returns
        -------
        q : Tensor
            Queries
        k : Tensor
            Keys
        v : Tensor
            Values
        """
        q = self.q_linear(inputs)

        *leading, items, last = q.shape
        hidden_channels = last // self.num_heads
        q = q.reshape(*leading, items, self.num_heads, hidden_channels)
        q = q.movedim(-2, -3)

        k = self.k_linear(inputs)[
            ..., None, :, :
        ]  # (..., head=1, item, hidden_channels)
        v = self.v_linear(inputs)[..., None, :, :]
        return q, k, v


class BaselineSelfAttention(nn.Module):
    """Baseline self-attention layer.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of input channels.
    hidden_channels : int
        Number of hidden channels = size of query, key, and value.
    attention
    num_heads : int
        Number of attention heads.
    multi_query : bool
        Use multi-query attention instead of multi-head attention.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: int,
        attention,
        num_heads: int = 8,
        multi_query: bool = True,
        dropout_prob=None,
    ) -> None:
        super().__init__()

        # Store settings
        self.num_heads = num_heads
        self.hidden_channels = hidden_channels

        self.attention = attention

        # Linear maps
        qkv_class = MultiQueryQKVLinear if multi_query else MultiHeadQKVLinear
        self.qkv_linear = qkv_class(in_channels, hidden_channels, num_heads)
        self.out_linear = nn.Linear(hidden_channels * num_heads, out_channels)

        if dropout_prob is not None:
            self.dropout = nn.Dropout(dropout_prob)
        else:
            self.dropout = None

    def forward(
        self,
        inputs: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        is_causal: bool = False,
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        inputs : Tensor
            Input data
        attention_mask : None or Tensor or xformers.ops.AttentionBias
            Optional attention mask

        Returns
        -------
        outputs : Tensor
            Outputs
        """
        q, k, v = self.qkv_linear(
            inputs
        )  # each: (..., num_heads, num_items, num_channels)

        # Attention layer
        h = self.attention(
            q.contiguous(),
            k.expand_as(q).contiguous(),
            v.expand_as(q),
            attn_mask=attention_mask,
            is_causal=is_causal,
        )

        # Concatenate heads and transform linearly
        *leading, num_heads, num_items, hidden_channels = h.shape
        h = h.permute(*range(len(leading)), -2, -3, -1)
        h = h.reshape(*leading, num_items, num_heads * hidden_channels)

        outputs = self.out_linear(h)  # (..., num_items, out_channels)

        if self.dropout is not None:
            outputs = self.dropout(outputs)

        return outputs


class BaselineTransformerBlock(nn.Module):
    """Baseline transformer block.

    Inputs are first processed by a block consisting of LayerNorm, multi-head self-attention, and
    residual connection. Then the data is processed by a block consisting of another LayerNorm, an
    item-wise two-layer MLP with GeLU activations, and another residual connection.

    Parameters
    ----------
    channels : int
        Number of input and output channels.
    attention
    num_heads : int
        Number of attention heads.
    increase_hidden_channels : int
        Factor by which the key, query, and value size is increased over the default value of
        hidden_channels / num_heads.
    multi_query : bool
        Use multi-query attention instead of multi-head attention.
    """

    def __init__(
        self,
        channels,
        attention,
        num_heads: int = 8,
        increase_hidden_channels=1,
        multi_query: bool = True,
        mlp_factor: int = 2,
        dropout_prob=None,
    ) -> None:
        super().__init__()

        self.norm = BaselineLayerNorm()

        hidden_channels = channels // num_heads * increase_hidden_channels

        self.attention = BaselineSelfAttention(
            channels,
            channels,
            hidden_channels,
            attention,
            num_heads=num_heads,
            multi_query=multi_query,
            dropout_prob=dropout_prob,
        )

        self.mlp = nn.Sequential(
            nn.Linear(channels, mlp_factor * channels),
            nn.Dropout(dropout_prob) if dropout_prob is not None else nn.Identity(),
            nn.GELU(),
            nn.Linear(mlp_factor * channels, channels),
            nn.Dropout(dropout_prob) if dropout_prob is not None else nn.Identity(),
        )

    def forward(
        self, inputs: torch.Tensor, attention_mask=None, is_causal=False
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        inputs : Tensor
            Input data
        attention_mask : None or Tensor or xformers.ops.AttentionBias
            Optional attention mask

        Returns
        -------
        outputs : Tensor
            Outputs
        """

        # Residual attention
        h = self.norm(inputs)
        h = self.attention(h, attention_mask=attention_mask, is_causal=is_causal)
        outputs = inputs + h

        # Residual MLP
        h = self.norm(outputs)
        h = self.mlp(h)
        outputs = outputs + h

        return outputs


class Transformer(nn.Module):
    """Baseline transformer.

    Combines num_blocks transformer blocks, each consisting of multi-head self-attention layers, an
    MLP, residual connections, and normalization layers.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    attn_reps : str
        Representation of each attention head.
    out_channels : int
        Number of output channels.
    num_blocks : int
        Number of transformer blocks.
    num_heads : int
        Number of attention heads.
    increase_hidden_channels : int
        Factor by which the key, query, and value size is increased over the default value of
        hidden_channels / num_heads.
    multi_query : bool
        Use multi-query attention instead of multi-head attention.
    """

    def __init__(
        self,
        in_channels: int,
        attn_reps: str,
        out_channels: int,
        num_blocks: int,
        num_heads: int,
        checkpoint_blocks: bool = False,
        increase_hidden_channels=1,
        mlp_factor: int = 2,
        multi_query: bool = False,
        dropout_prob=None,
    ) -> None:
        super().__init__()
        attn_reps = TensorReps(attn_reps)
        self.hidden_channels = attn_reps.dim * num_heads
        self.checkpoint_blocks = checkpoint_blocks
        self.attention = LLoCaAttention(attn_reps, num_heads)

        self.linear_in = nn.Linear(in_channels, self.hidden_channels)
        self.blocks = nn.ModuleList(
            [
                BaselineTransformerBlock(
                    self.hidden_channels,
                    attention=self.attention,
                    num_heads=num_heads,
                    increase_hidden_channels=increase_hidden_channels,
                    mlp_factor=mlp_factor,
                    multi_query=multi_query,
                    dropout_prob=dropout_prob,
                )
                for _ in range(num_blocks)
            ]
        )
        self.linear_out = nn.Linear(self.hidden_channels, out_channels)

    def forward(
        self, inputs: torch.Tensor, frames, attention_mask=None, is_causal=False
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        inputs : Tensor
            Input data with shape (..., num_items, in_channels)
        frames : Frames
            Local frames used for invariant particle attention
        attention_mask : None or Tensor or xformers.ops.AttentionBias
            Optional attention mask
        is_causal: bool

        Returns
        -------
        outputs : Tensor
            Outputs with shape (..., num_items, out_channels)
        """
        self.attention.prepare_frames(frames)

        h = self.linear_in(inputs)
        for block in self.blocks:
            if self.checkpoint_blocks:
                fn = partial(block, attention_mask=attention_mask, is_causal=is_causal)
                h = checkpoint(fn, h)
            else:
                h = block(h, attention_mask=attention_mask, is_causal=is_causal)
        outputs = self.linear_out(h)
        return outputs



from torch import nn


class EquiVectors(nn.Module):
    """Abstract class for equivariantly predicting vectors
    based on fourmomenta and scalars"""

    def __init__(self):
        super().__init__()

    def forward(self, fourmomenta, scalars, *args, **kwargs):
        raise NotImplementedError



import torch

from .lorentz import (
    lorentz_inner,
    lorentz_squarednorm,
    lorentz_metric,
    lorentz_cross,
)


def orthogonalize_4d(vecs, use_float64=True, return_reg=False, **kwargs):
    """High-level wrapper for orthogonalization of three Lorentz vectors.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of three Lorentz vectors of shape (..., 4).
    use_float64 : bool
        If True, use float64 for numerical stability during orthogonalization.
    return_reg : bool
        If True, return a tuple with the orthogonalized vectors and the number of
        regularized vectors for lightlike and coplanar cases.
    kwargs : dict
        Additional keyword arguments passed to the orthogonalization function.

    Returns
    -------
    trafo : torch.Tensor
        Lorentz transformation of shape (..., 4, 4) that orthogonalizes the input vectors.
        The first vector is guaranteed to be timelike.
    reg_lightlike : int
        Number of vectors that were regularized due to being lightlike.
    reg_coplanar : int
        Number of vectors that were regularized due to coplanarity.
    """
    if use_float64:
        original_dtype = vecs[0].dtype
        vecs = [v.to(torch.float64) for v in vecs]

    out = orthogonalize_wrapper_4d(vecs, return_reg=return_reg, **kwargs)
    if return_reg:
        orthogonal_vecs, *reg = out
    else:
        orthogonal_vecs = out
    trafo = torch.stack(orthogonal_vecs, dim=-2)

    trafo = timelike_first(trafo)
    metric = lorentz_metric(trafo.shape[:-2], device=trafo.device, dtype=trafo.dtype)
    trafo = metric @ trafo @ metric
    if use_float64:
        trafo = trafo.to(original_dtype)
    return (trafo, *reg) if return_reg else trafo


def orthogonalize_wrapper_4d(
    vecs,
    method="gramschmidt",
    eps_norm=1e-15,
    eps_reg_coplanar=1e-10,
    eps_reg_lightlike=1e-10,
    return_reg=False,
):
    """Wrapper for orthogonalization of Minkowski vectors.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of three Lorentz vectors of shape (..., 4).
    method : str
        Method for orthogonalization. Options are "cross" and "gramschmidt".
    eps_norm : float
        Numerical regularization for the normalization of the vectors.
    eps_reg_coplanar : float
        Controls the scale of the regularization for coplanar vectors.
        eps_reg_coplanar**2 defines the selection threshold.
    eps_reg_lightlike : float
        Controls the scale of the regularization for lightlike vectors.
        eps_reg_lightlike**2 defines the selection threshold.
    return_reg : bool
        If True, return a tuple with the orthogonalized vectors and the number of
        regularized vectors for lightlike and coplanar cases.

    Returns
    -------
    orthogonal_vecs : list of torch.Tensor
        List of orthogonalized Lorentz vectors of shape (..., 4).
    reg_lightlike : int
        Number of vectors that were regularized due to being lightlike.
    reg_coplanar : int
        Number of vectors that were regularized due to coplanarity.
    """
    assert len(vecs) == 3
    assert all(v.shape == vecs[0].shape for v in vecs)

    vecs, reg_lightlike = regularize_lightlike(vecs, eps_reg_lightlike)
    vecs, reg_coplanar = regularize_coplanar(vecs, eps_reg_coplanar)

    if method == "cross":
        trafo = orthogonalize_cross(vecs, eps_norm)
    elif method == "gramschmidt":
        trafo = orthogonalize_gramschmidt(vecs, eps_norm)
    else:
        raise ValueError(f"Orthogonalization method {method} not implemented")

    return (trafo, reg_lightlike, reg_coplanar) if return_reg else trafo


def orthogonalize_gramschmidt(vecs, eps_norm=1e-15):
    """Gram-Schmidt orthogonalization algorithm for Minkowski vectors.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of Lorentz vectors of shape (..., 4).
    eps_norm : float
        Small value to avoid division by zero during normalization.

    Returns
    -------
    orthogonal_vecs : list of torch.Tensor
        List of orthogonalized Lorentz vectors of shape (..., 4).
    """
    vecs = [normalize_4d(v, eps_norm) for v in vecs]

    v_nexts = [v for v in vecs]
    orthogonal_vecs = [vecs[0]]
    for i in range(1, len(vecs)):
        for k in range(i, len(vecs)):
            v_inner = lorentz_inner(v_nexts[k], orthogonal_vecs[i - 1]).unsqueeze(-1)
            v_norm = lorentz_squarednorm(orthogonal_vecs[i - 1]).unsqueeze(-1)
            v_nexts[k] = v_nexts[k] - orthogonal_vecs[i - 1] * v_inner / (
                v_norm + eps_norm
            )
        orthogonal_vecs.append(normalize_4d(v_nexts[i], eps_norm))
    last_vec = normalize_4d(lorentz_cross(*orthogonal_vecs), eps_norm)
    orthogonal_vecs.append(last_vec)

    return orthogonal_vecs


def orthogonalize_cross(vecs, eps_norm=1e-15):
    """Orthogonalization algorithm using repeated cross products.
    This approach gives the same result as orthogonalize_gramschmidt,
    but we find empirically that the Gram-Schmidt approach is more stable.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of Lorentz vectors of shape (..., 4).
    eps_norm : float
        Small value to avoid division by zero during normalization.

    Returns
    -------
    orthogonal_vecs : list of torch.Tensor
        List of orthogonalized Lorentz vectors of shape (..., 4).
    """
    vecs = [normalize_4d(v, eps_norm) for v in vecs]

    orthogonal_vecs = [vecs[0]]
    for i in range(1, len(vecs) + 1):
        v_next = lorentz_cross(*orthogonal_vecs, *vecs[i:])
        assert torch.isfinite(v_next).all()
        orthogonal_vecs.append(normalize_4d(v_next, eps_norm))

    return orthogonal_vecs


def timelike_first(trafo):
    """Reorder the Lorentz transformation such that the first vector is timelike.
    This is necessary to ensure that the resulting Lorentz transformation has the
    correct metric signature (1, -1, -1, -1). Note that this step can be skipped
    if the first vector is already timelike.

    Parameters
    ----------
    trafo : torch.Tensor
        Lorentz transformation of shape (..., 4, 4) where the last two dimensions
        represent the transformation matrix.

    Returns
    -------
    trafo : torch.Tensor
        Lorentz transformation of shape (..., 4, 4) with the first vector being timelike.
    """
    vecs = [trafo[..., i, :] for i in range(4)]
    norm = torch.stack([lorentz_squarednorm(v) for v in vecs], dim=-1)
    pos_norm = norm > 0
    num_pos_norm = pos_norm.sum(dim=-1)
    assert (
        num_pos_norm == 1
    ).all(), f"Don't always have exactly 1 timelike vector: {(num_pos_norm==0).sum().item()} (#0), {(num_pos_norm==1).sum().item()} (#1), {(num_pos_norm==2).sum().item()} (#2), {(num_pos_norm==3).sum().item()} (#3)"
    old_trafo = trafo.clone()
    trafo[..., 0, :] = old_trafo[pos_norm].view(*trafo.shape[:-2], 4)
    trafo[..., 1:, :] = old_trafo[~pos_norm].view(*trafo.shape[:-2], 3, 4)
    return trafo


def regularize_lightlike(vecs, eps_reg_lightlike=1e-10):
    """If the Minkowski norm of a vector is close to zero,
    it is lightlike. In this case, we add a bit of noise to the vector
    to break the degeneracy and ensure that the orthogonalization works.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of Lorentz vectors of shape (..., 4).
    eps_reg_lightlike : float
        Small value to control the scale of the regularization for lightlike vectors.

    Returns
    -------
    vecs_reg : list of torch.Tensor
        List of Lorentz vectors of shape (..., 4) with regularization applied.
    reg_lightlike : int
        Number of vectors that were regularized due to being lightlike.
    """
    vecs_reg = []
    masks = []
    for v in vecs:
        inners = lorentz_inner(v, v)
        mask = inners.abs() < eps_reg_lightlike**2
        v_reg = v + eps_reg_lightlike * torch.randn_like(v) * mask.unsqueeze(-1)
        masks.append(mask)
        vecs_reg.append(v_reg)

    reg_lightlike = torch.stack(masks).any(dim=-1).sum().item()
    return vecs_reg, reg_lightlike


def regularize_coplanar(vecs, eps_reg_coplanar=1e-10):
    """If the cross product of three vectors is close to zero,
    they are coplanar. In this case, we add a bit of noise to each vector
    to break the degeneracy and ensure that the orthogonalization works.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of three Lorentz vectors of shape (..., 4).
    eps_reg_coplanar : float
        Small value to control the scale of the regularization for coplanar vectors.

    Returns
    -------
    vecs_reg : list of torch.Tensor
        List of three Lorentz vectors of shape (..., 4) with regularization applied.
    reg_coplanar : int
        Number of vectors that were regularized due to coplanarity.
    """
    assert len(vecs) == 3
    cross_norm = lorentz_squarednorm(lorentz_cross(*vecs))
    mask = cross_norm.abs() < eps_reg_coplanar**2

    vecs_reg = []
    for v in vecs:
        v_reg = v + eps_reg_coplanar * torch.randn_like(v) * mask.unsqueeze(-1)
        vecs_reg.append(v_reg)

    reg_coplanar = mask.sum().item()
    return vecs_reg, reg_coplanar


def normalize_4d(v, eps=1e-15):
    """Normalize a Lorentz vector by the absolute value of the Minkowski norm.
    Note that this norm can be close to zero.

    Parameters
    ----------
    v : torch.Tensor
        Lorentz vector of shape (..., 4).
    eps : float
        Small value to avoid division by zero.

    Returns
    -------
    torch.Tensor
        Normalized Lorentz vector of shape (..., 4).
    """
    norm = lorentz_squarednorm(v).unsqueeze(-1)
    norm = norm.abs().sqrt()  # could also multiply by torch.sign(norm)
    return v / (norm + eps)



# Non-equivariant local frames, i.e. non-equivariant networks and data augmentation
import torch

from .frames import Frames
from ..utils.rand_transforms import (
    rand_lorentz,
    rand_rotation,
    rand_xyrotation,
    rand_ztransform,
)
from ..utils.polar_decomposition import restframe_boost


class FramesPredictor(torch.nn.Module):
    def __init__(self, is_global=False, is_identity=False):
        super().__init__()
        self.is_global = is_global
        self.is_identity = is_identity

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        raise NotImplementedError


class IdentityFrames(FramesPredictor):
    """Identity local frames, corresponding to non-equivariant networks"""

    def __init__(self):
        super().__init__(is_global=True, is_identity=True)

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        frames = Frames(
            is_identity=True,
            device=fourmomenta.device,
            dtype=fourmomenta.dtype,
            shape=fourmomenta.shape[:-1],
        )

        return (frames, {}) if return_tracker else frames

    def __repr__(self):
        return "IdentityFrames()"


class RandomFrames(FramesPredictor):
    """Randomly generates a local frame for the whole batch,
    corresponding to data augmentation."""

    def __init__(
        self,
        transform_type="lorentz",
        is_global=True,
        std_eta=0.1,
        n_max_std_eta=3.0,
    ):
        super().__init__(is_global=is_global)
        self.is_global = is_global
        self.std_eta = std_eta
        self.transform_type = transform_type
        self.n_max_std_eta = n_max_std_eta

    def transform(self, shape, device, dtype):
        if self.transform_type == "lorentz":
            return rand_lorentz(
                shape,
                std_eta=self.std_eta,
                n_max_std_eta=self.n_max_std_eta,
                device=device,
                dtype=dtype,
            )
        elif self.transform_type == "rotation":
            return rand_rotation(shape, device=device, dtype=dtype)
        elif self.transform_type == "xyrotation":
            return rand_xyrotation(shape, device=device, dtype=dtype)
        elif self.transform_type == "ztransform":
            return rand_ztransform(
                shape,
                std_eta=self.std_eta,
                n_max_std_eta=self.n_max_std_eta,
                device=device,
                dtype=dtype,
            )
        else:
            raise ValueError(
                f"Transformation type {self.transform_type} not implemented"
            )

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        if not self.training:
            frames = Frames(
                is_identity=True,
                shape=fourmomenta.shape[:-1],
                device=fourmomenta.device,
                dtype=fourmomenta.dtype,
            )
            return (frames, {}) if return_tracker else frames

        shape = (
            fourmomenta.shape[:-2] + (1,) if self.is_global else fourmomenta.shape[:-1]
        )
        matrix = self.transform(
            shape, device=fourmomenta.device, dtype=fourmomenta.dtype
        )
        matrix = matrix.expand(*fourmomenta.shape[:-1], 4, 4)

        frames = Frames(
            is_global=self.is_global,
            matrices=matrix,
            device=fourmomenta.device,
            dtype=fourmomenta.dtype,
        )
        return (frames, {}) if return_tracker else frames

    def __repr__(self):
        string = f"RandomFrames(transform_type={self.transform_type}, is_global={self.is_global}"
        if self.transform_type in ["lorentz", "ztransform, general_lorentz"]:
            string += f", std_eta={self.std_eta}"
            string += f", n_max_std_eta={self.n_max_std_eta}"
        string += ")"
        return string


class COMRandomFrames(RandomFrames):
    """Modifies the forward function of RandomFrames such that
    an additional boost is applied to the whole event.

    Only applicable to amplitude regression, the boost changes
    the reference frame to the center of mass of the incoming particles.
    """

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        if not self.training:
            frames = Frames(
                is_identity=True,
                shape=fourmomenta.shape[:-1],
                device=fourmomenta.device,
                dtype=fourmomenta.dtype,
            )
            return (frames, {}) if return_tracker else frames

        shape = (
            fourmomenta.shape[:-2] + (1,) if self.is_global else fourmomenta.shape[:-1]
        )
        matrix = self.transform(
            shape, device=fourmomenta.device, dtype=fourmomenta.dtype
        )

        # hardcoded for amplitudes
        reference_vector = fourmomenta[..., :2, :].sum(dim=-2, keepdims=True)
        reference_boost = restframe_boost(reference_vector)[..., :4, :4]

        matrix = torch.einsum("...ij,...jk->...ik", matrix, reference_boost)
        matrix = matrix.expand(*fourmomenta.shape[:-1], 4, 4)

        frames = Frames(
            is_global=self.is_global,
            matrices=matrix,
            device=fourmomenta.device,
            dtype=fourmomenta.dtype,
        )
        return (frames, {}) if return_tracker else frames

    def __repr__(self):
        string = f"COMFrames(transform_type={self.transform_type}, is_global={self.is_global}"
        if self.transform_type in ["lorentz", "ztransform", "general_lorentz"]:
            string += f", std_eta={self.std_eta}"
            string += f", n_max_std_eta={self.n_max_std_eta}"
        string += ")"
        return string



from typing import Tuple, List


class TensorRep(Tuple):
    """Individual tensor representation"""

    def __new__(cls, order, parity):
        """Create a tensor representation based on its order and parity.

        Parameters
        ----------
        order: int
            The order of the tensor representation.
        parity: int
            The parity of the tensor representation. Can be -1 or 1
        """
        assert isinstance(order, int) and order >= 0, order
        assert parity in [-1, 1]
        return super().__new__(cls, (order, parity))

    def __deepcopy__(self, memo):
        return self

    @property
    def order(self) -> int:
        """The order of the tensor."""
        return self[0]

    @property
    def parity(self):
        """The parity of the tensor."""
        return self[1]

    def __repr__(self):
        """Returns a string representation of the tensor."""
        return f"{self.order}{'n' if self.parity == 1 else 'p'}"


class _TensorMulRep(Tuple):
    """Direct product of similar tensor representations"""

    def __new__(cls, mul, rep):
        """Create a tensor multiplication representation.
        This represents a direct product of a tensor representation with a multiplier.

        Parameters
        ----------
        mul: int
            The multiplier for the tensor representation.
        rep: TensorRep
            The tensor representation.
        """

        assert isinstance(mul, int), "mul must be an integer"
        assert isinstance(rep, TensorRep), "rep must be an instance of TensorRep"

        return super().__new__(cls, (mul, rep))

    def __deepcopy__(self, memo):
        return self

    @property
    def mul(self):
        """The multiplier for the tensor representation."""
        return self[0]

    @property
    def rep(self):
        """The tensor representation."""
        return self[1]

    @property
    def dim(self):
        """The dimension of the tensor multiplication."""
        if self.rep.order == 0:
            return 1 * self.mul
        else:
            return (4**self.rep.order) * self.mul

    def __repr__(self):
        """Returns a string representation of the tensor multiplication."""
        return f"{self.mul}x{self.rep}"


class TensorReps(Tuple):
    """Direct product of potentially different tensor representations"""

    def __new__(cls, input, simplify=True):
        """Create a tensor representation based on the input.
        This can be a string representation or a list of _TensorMulRep instances.

        Parameters
        ----------
        input: Union[TensorReps, str, List[_TensorMulRep]]
            The input to initialize the tensor reps.
            If `input` is a string, it is parsed to extract the tensor representations.
            If `input` is a list, it should contain instances of `_TensorMulRep`.
        simplify: bool
            Whether to simplify the tensor reps after initialization. Default is True.
        """
        if isinstance(input, TensorReps):
            tensor_reps = input
        elif isinstance(input, List):
            assert all(isinstance(x, _TensorMulRep) for x in input)
            tensor_reps = input
        elif isinstance(input, str):
            try:
                tensor_reps = parse_tensorreps_string(input)
            except Exception:
                raise ValueError(f"Invalid tensor_reps string {input}")
        else:
            raise ValueError(f"Invalid input: {input} is of type {type(input)}")

        ret = super().__new__(cls, tensor_reps)
        if simplify:
            return ret.simplify()
        return ret

    def __repr__(self) -> str:
        """Returns a string representation of the tensor reps."""
        return "+".join(f"{mul_ir}" for mul_ir in self)

    def __deepcopy__(self, memo):
        return self

    @property
    def dim(self) -> int:
        """The total dimension of the tensor reps."""
        return sum(mul_ir.dim for mul_ir in self)

    @property
    def max_rep(self) -> TensorRep:
        """The tensor irrep with the highest order."""
        return max(self, key=lambda x: x.rep.order)

    @property
    def mul_without_scalars(self) -> int:
        """The total multiplier of the tensor reps without the scalars."""
        return sum(mul_ir.mul for mul_ir in self if mul_ir.rep.order != 0)

    @property
    def mul_scalars(self) -> int:
        """The total multiplier of the tensor reps scalars."""
        return sum(mul_ir.mul for mul_ir in self if mul_ir.rep.order == 0)

    @property
    def mul(self) -> int:
        """The total multiplier of the tensor reps."""
        return sum(mul_ir.mul for mul_ir in self)

    @property
    def reps(self) -> set:
        """The set of tensor reps."""
        return {rep for _, rep in self}

    @property
    def is_sorted(self) -> bool:
        """Whether the tensor reps are sorted by the order of the reps."""
        if len(self) <= 1:
            return True
        else:
            return all(
                self[i].rep.order <= self[i + 1].rep.order
                for i in range(len(self[:-1]))
            )

    def __add__(self, tensor_reps, simplify=True) -> "TensorReps":
        """Adds tensor reps to the current tensor reps."""
        tensor_reps = TensorReps(tensor_reps)
        return TensorReps(super().__add__(tensor_reps), simplify=simplify)

    def sort(self):
        """Sorts the tensor reps by the order of the reps."""
        return TensorReps(sorted(self, key=lambda x: x.rep.order))

    def simplify(self) -> "TensorReps":
        """Simplifies the tensor reps by combining the same reps."""
        if not self.is_sorted:
            self = self.sort()

        out = []
        for mul_rep in self:
            mul, rep = mul_rep
            if len(out) > 0 and out[-1].rep == rep:
                # same rep -> extend mul of previous rep
                out[-1] = _TensorMulRep(out[-1].mul + mul, rep)
            elif mul > 0:
                # different rep and mul>0 -> create new entry
                out.append(mul_rep)

        return TensorReps(out, simplify=False)

    def is_simplified(self):
        """Check if the TensorReps is simlified."""
        if not self.is_sorted:
            return False

        return all(
            self[i].rep.order != self[i + 1].rep.order for i in range(len(self[:-1]))
        )


def parse_tensorreps_string(input):
    """Parse a string representation of tensor representations into a list of _TensorMulRep instances.

    Parameters
    ----------
    input: str
        The string representation of tensor representations, e.g. "2x2n+3x1p+1x0n".

    Returns
    -------
    List[_TensorMulRep]
        A list of _TensorMulRep instances representing the tensor representations.
    """
    out = []
    input = input.replace(" ", "")  # remove whitespace
    input_list = input.split("+")  # split into single _TensorMulRep
    for i, rep in enumerate(input_list):
        if rep[-1] == "n":
            parity = 1
            rep = rep[:-1]
        elif rep[-1] == "p":
            parity = -1
            rep = rep[:-1]
        else:
            raise ValueError(
                f"Invalid last character (=parity) in tensorreps string {input_list[i][-1]}, should be either 'n' or 'p'"
            )

        mul, order = rep.split("x")
        mul, order = int(mul), int(order)
        out.append(_TensorMulRep(mul, TensorRep(order, parity)))
    return out



import torch

from .orthogonalize_3d import orthogonalize_3d
from .lorentz import lorentz_squarednorm


def restframe_boost(fourmomenta):
    """Construct a Lorentz transformation that boosts the given four-momenta
    into their rest frame. This transformation alone does not have the
    frames transformation properties, because it does not include a rotation.

    Parameters
    ----------
    fourmomenta : torch.Tensor
        Tensor of shape (..., 4) representing the four-momenta.

    Returns
    -------
    trafo : torch.Tensor
        Tensor of shape (..., 4, 4) representing the Lorentz transformation
        that boosts the four-momenta into their rest frame.
    """
    assert (
        lorentz_squarednorm(fourmomenta) > 0
    ).all(), "Trying to boost spacelike vectors into their restframe (not possible). Consider changing the nonlinearity in equivectors."

    beta = fourmomenta[..., 1:] / fourmomenta[..., [0]].clamp(min=1e-10)
    beta2 = (beta**2).sum(dim=-1, keepdim=True)
    gamma = 1 / (1 - beta2).clamp(min=1e-10).sqrt()

    # prepare entries of the trafo
    boost = -gamma * beta
    eye = torch.eye(3, device=fourmomenta.device, dtype=fourmomenta.dtype)
    eye = eye.view(*(1,) * len(fourmomenta.shape[:-1]), 3, 3).repeat(
        *fourmomenta.shape[:-1], 1, 1
    )
    rot = eye + (
        (gamma[..., None] - 1)
        * beta[..., None]
        * beta[..., None, :]
        / beta2[..., None].clamp(min=1e-10)
    )

    # put trafo together
    trafo = torch.empty(
        *fourmomenta.shape[:-1],
        4,
        4,
        device=fourmomenta.device,
        dtype=fourmomenta.dtype
    )
    trafo[..., 0, 0] = gamma[..., 0]
    trafo[..., 1:, 1:] = rot
    trafo[..., 0, 1:] = boost
    trafo[..., 1:, 0] = boost
    assert torch.isfinite(trafo).all()
    return trafo


def polar_decomposition(
    fourmomenta, references, use_float64=True, return_reg=False, **kwargs
):
    """Construct a Lorentz transformation as a polar decomposition of a
    boost and a rotation.

    Parameters
    ----------
    fourmomenta : torch.Tensor
        Tensor of shape (..., 4) representing the four-momenta that define the rest frames.
    references : List[torch.Tensor]
        List of two tensors of shape (..., 4) representing the reference four-momenta
        to construct the rotation.
    use_float64 : bool
        If True, use float64 for calculations to avoid numerical issues.
    return_reg : bool
        If True, return a tuple with the Lorentz transformation and regularization information.
    kwargs : dict

    Returns
    -------
    trafo : torch.Tensor
        Tensor of shape (..., 4, 4) representing the Lorentz transformation.
    reg_collinear : torch.Tensor, optional
        Tensor indicating if the references were collinear (only returned if `return_reg` is True).
    """
    assert len(references) == 2
    assert all(r.shape == fourmomenta.shape for r in references)

    if use_float64:
        original_dtype = fourmomenta.dtype
        fourmomenta = fourmomenta.to(torch.float64)
        references = [r.to(torch.float64) for r in references]

    # construct rest frame transformation
    boost = restframe_boost(fourmomenta)

    # references go into rest frame
    ref_rest = [torch.einsum("...ij,...j->...i", boost, v) for v in references]

    # construct rotation
    ref3_rest = [r[..., 1:] for r in ref_rest]
    out = orthogonalize_3d(ref3_rest, return_reg=return_reg, **kwargs)
    if return_reg:
        orthogonal_vec3, reg_collinear = out
    else:
        orthogonal_vec3 = out
    rotation = torch.zeros_like(boost)
    rotation[..., 0, 0] = 1
    rotation[..., 1:, 1:] = torch.stack(orthogonal_vec3, dim=-2)

    # combine rotation and boost
    trafo = torch.einsum("...ij,...jk->...ik", rotation, boost)
    if use_float64:
        trafo = trafo.to(original_dtype)
    assert torch.isfinite(trafo).all()
    return (trafo, reg_collinear) if return_reg else trafo



import torch
import pytest
from tests.constants import TOLERANCES, REPS

from lloca.framesnet.frames import Frames, InverseFrames, ChangeOfFrames
from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.utils.rand_transforms import rand_lorentz


@pytest.mark.parametrize("batch_dims", [[1000]])
@pytest.mark.parametrize("reps", REPS)
def test_equivariance(batch_dims, reps):
    dtype = torch.float64

    reps = TensorReps(reps)
    trafo = TensorRepsTransform(TensorReps(reps))

    transform = rand_lorentz(batch_dims, dtype=dtype)
    frames = Frames(transform)

    x = torch.randn(*batch_dims, reps.dim, dtype=dtype)

    # manual transform
    transform_direct = torch.einsum(
        "...ij,...jk->...ik", frames.matrices, InverseFrames(frames).matrices
    )
    change_frames1 = Frames(transform_direct)
    x_prime1 = trafo(x, change_frames1)
    torch.testing.assert_close(x, x_prime1, **TOLERANCES)

    # all-in-one transform
    change_frames2 = ChangeOfFrames(frames, frames)
    x_prime2 = trafo(x, change_frames2)
    torch.testing.assert_close(x, x_prime2, **TOLERANCES)



import torch
import pytest
from tests.constants import TOLERANCES, LOGM2_MEAN_STD, REPS, FRAMES_PREDICTOR
from tests.helpers import sample_particle, equivectors_builder

from lloca.backbone.transformer import Transformer
from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.utils.rand_transforms import rand_lorentz
from lloca.framesnet.frames import InverseFrames


@pytest.mark.parametrize("FramesPredictor", FRAMES_PREDICTOR)
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("num_heads", [1, 4])
@pytest.mark.parametrize("num_blocks", [0, 1, 2])
@pytest.mark.parametrize("attn_reps", REPS)
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_transformer_invariance_equivariance(
    FramesPredictor,
    batch_dims,
    num_heads,
    num_blocks,
    logm2_std,
    logm2_mean,
    attn_reps,
):
    dtype = torch.float64

    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # define edgeconv
    in_reps = TensorReps("1x1n")
    trafo = TensorRepsTransform(TensorReps(in_reps))
    net = Transformer(
        in_channels=in_reps.dim,
        attn_reps=attn_reps,
        out_channels=in_reps.dim,
        num_blocks=num_blocks,
        num_heads=num_heads,
    ).to(dtype=dtype)

    # get global transformation
    random = rand_lorentz([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    frames = call_predictor(fm)
    fm_local = trafo(fm, frames)

    # global - edgeconv
    fm_transformed = torch.einsum("...ij,...j->...i", random, fm)
    frames_transformed = call_predictor(fm_transformed)
    fm_tr_local = trafo(fm_transformed, frames_transformed)
    fm_tr_prime_local = net(fm_tr_local, frames_transformed)
    # back to global frame
    fm_tr_prime_global = trafo(fm_tr_prime_local, InverseFrames(frames_transformed))

    # edgeconv - global
    fm_prime_local = net(fm_local, frames)
    # back to global
    fm_prime_global = trafo(fm_prime_local, InverseFrames(frames))
    fm_prime_tr_global = torch.einsum("...ij,...j->...i", random, fm_prime_global)

    # test equivariance of outputs
    torch.testing.assert_close(fm_tr_prime_global, fm_prime_tr_global, **TOLERANCES)



import pytest
import torch
from tests.constants import TOLERANCES, BATCH_DIMS

from lloca.utils.rand_transforms import rand_lorentz
from lloca.framesnet.frames import Frames
from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform


@pytest.mark.parametrize("use_naive", [True, False])
@pytest.mark.parametrize("batch_dim", BATCH_DIMS)
def test_tensorreps(batch_dim, use_naive):
    frames = rand_lorentz(batch_dim)
    frames = Frames(frames)

    rep = "5x0n+3x1n+3x2n+5x3n"
    rep = TensorReps(rep)
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    tensor_reps_transform(coeffs, frames)

    # 0n test
    rep = TensorReps("5x0n")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    torch.testing.assert_close(transformed_coeffs, coeffs, **TOLERANCES)

    # 0p test
    rep = TensorReps("5x0p")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    torch.testing.assert_close(
        transformed_coeffs, coeffs * frames.det[..., None], **TOLERANCES
    )

    # 1n test
    rep = TensorReps("5x1n")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    torch.testing.assert_close(
        transformed_coeffs.reshape(*batch_dim, rep.max_rep.mul, 4),
        torch.matmul(
            coeffs.reshape(*batch_dim, rep.max_rep.mul, 4),
            frames.matrices.transpose(-1, -2),
        ),
        **TOLERANCES,
    )

    # 1p test
    rep = TensorReps("5x1p")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    torch.testing.assert_close(
        transformed_coeffs.reshape(*batch_dim, rep.max_rep.mul, 4),
        torch.matmul(
            coeffs.reshape(*batch_dim, rep.max_rep.mul, 4),
            frames.matrices.transpose(-1, -2),
        )
        * frames.det[..., None, None],
        **TOLERANCES,
    )

    # 2n test
    rep = TensorReps("5x2n")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    naive_trafo = torch.einsum(
        "...ij, ...lm, ...cjm -> ...cil",
        frames.matrices,
        frames.matrices,
        coeffs.reshape(*batch_dim, rep.max_rep.mul, 4, 4),
    )
    torch.testing.assert_close(
        transformed_coeffs.reshape(*batch_dim, rep.max_rep.mul, 4, 4),
        naive_trafo,
        **TOLERANCES,
    )

    # 2p test
    rep = TensorReps("5x2p")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    naive_trafo = torch.einsum(
        "...ij, ...lm, ...cjm -> ...cil",
        frames.matrices,
        frames.matrices,
        coeffs.reshape(*batch_dim, rep.max_rep.mul, 4, 4),
    )
    naive_trafo *= frames.det[..., None, None, None]
    torch.testing.assert_close(
        transformed_coeffs.reshape(*batch_dim, rep.max_rep.mul, 4, 4),
        naive_trafo,
        **TOLERANCES,
    )

    # 1x0n+1x2n test
    rep = TensorReps("1x0n+1x2n")
    coeffs = torch.randn(batch_dim + [rep.dim])
    tensor_reps_transform = TensorRepsTransform(rep)
    transformed_coeffs = tensor_reps_transform(coeffs.clone(), frames)
    naive_trafo = torch.einsum(
        "...ij, ...lm, ...cjm -> ...cil",
        frames.matrices,
        frames.matrices,
        coeffs[..., 1:].reshape(*batch_dim, rep.max_rep.mul, 4, 4),
    )
    torch.testing.assert_close(
        transformed_coeffs[..., 1:].reshape(*batch_dim, rep.max_rep.mul, 4, 4),
        naive_trafo,
        **TOLERANCES,
    )
    torch.testing.assert_close(
        coeffs[..., :1], transformed_coeffs[..., :1], **TOLERANCES
    )



from math import prod
import torch
from torch import Tensor

from ..framesnet.frames import Frames, InverseFrames, LowerIndicesFrames
from ..reps.tensorreps import TensorReps
from ..reps.tensorreps_transform import TensorRepsTransform
from .attention_backends import get_attention_backend


class LLoCaAttention(torch.nn.Module):
    """"""

    def __init__(self, attn_reps, num_heads):
        super().__init__()
        self.transform = TensorRepsTransform(TensorReps(attn_reps))
        self.num_heads = num_heads

        self.frames = None
        self.inv_frames = None
        self.lower_inv_frames = None

    def prepare_frames(self, frames):
        """Prepare local frames for processing with LLoCa attention.
        For a single forward pass through the network, this method is
        called only once for efficiency.

        Parameters
        ----------
        frames: torch.tensor
            Local frames of reference for each particle of shape (..., N, 4, 4)
            where N is the number of particles.
        """
        self.frames = frames
        if not self.frames.is_global:
            # insert frames head dimension
            self.frames = self.frames.reshape(
                *frames.shape[:-3], 1, frames.shape[-3], 4, 4
            )
            self.frames = self.frames.repeat(
                *((1,) * len(frames.shape[:-3])), self.num_heads, 1, 1, 1
            )

            # create inv_frames and lower_inv_frames
            inv_frames = InverseFrames(self.frames)
            lower_inv_frames = LowerIndicesFrames(inv_frames)

            # qkv = (inv_frames, lower_inv_frames, inv_frames)
            # note that (lower_inv_frames, inv_frames, inv_frames) is equivalent
            self.frames_qkv = Frames(
                matrices=torch.stack(
                    [
                        inv_frames.matrices,
                        lower_inv_frames.matrices,
                        inv_frames.matrices,
                    ],
                    dim=0,
                ),
                is_identity=inv_frames.is_identity,
                is_global=inv_frames.is_global,
                det=torch.stack(
                    [inv_frames.det, lower_inv_frames.det, inv_frames.det], dim=0
                ),
                inv=torch.stack(
                    [inv_frames.inv, lower_inv_frames.inv, inv_frames.inv], dim=0
                ),
            )

            # flatten frames (preparation for tensorreps_transform)
            self.frames = self.frames.reshape(-1, 4, 4)
            self.frames_qkv = self.frames_qkv.reshape(-1, 4, 4)

    def forward(self, q_local, k_local, v_local, **attn_kwargs):
        """Execute LLoCa attention.

        Strategy
        1) Transform q, k, v into global frame
        2) Apply attention in global frame
        3) Transform output back into local frame

        Comments
        - dimensions: *dims (optional), H (head), N (particles), C (channels)
        - extension to cross-attention is trivial but we don't have this right now for convenience
          strategy: frames_q for queries (in contrast to frames=frames_kv)

        Parameters
        ----------
        q_local: torch.tensor
            Local queries of shape (*dims, H, N, C)
        k_local: torch.tensor
            Local keys of shape (*dims, H, N, C)
        v_local: torch.tensor
            Local values of shape (*dims, H, N, C)
        attn_kwargs: dict
            Optional arguments that are passed on to attention

        Returns
        -------
        out_local: torch.tensor
            Attention output in local frame of shape (*dims, H, N, C)
        """
        if self.frames.is_global:
            # fallback to standard attention for global frames
            return scaled_dot_product_attention(
                q_local,
                k_local,
                v_local,
                **attn_kwargs,
            )

        # check input shapes
        assert k_local.shape == v_local.shape == q_local.shape  # has to match perfectly
        assert 3 * prod(k_local.shape[:-1]) == self.frames_qkv.shape[-3]

        # transform q, k, v into global frame
        qkv_local = torch.stack([q_local, k_local, v_local], dim=0)
        qkv_global = self.transform(qkv_local, self.frames_qkv)
        q_global, k_global, v_global = torch.unbind(qkv_global, dim=0)

        # (B, H, N, C) format required for scaled_dot_product_attention
        shape_q, shape_k = q_global.shape, k_global.shape
        q_global = q_global.reshape(-1, *shape_q[-3:])
        k_global = k_global.reshape(-1, *shape_k[-3:])
        v_global = v_global.reshape(-1, *shape_k[-3:])

        # attention (in global frame)
        out_global = scaled_dot_product_attention(
            q_global,
            k_global,
            v_global,
            **attn_kwargs,
        )

        out_global = out_global.view(*shape_q)  # (*dims, H, N, C)

        # transform result back into local frame
        out_local = self.transform(out_global, self.frames)
        return out_local


def scaled_dot_product_attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    **attn_kwargs,
) -> Tensor:
    """Execute scaled dot-product attention.
    The attention backend is determined dynamically
    based on the ``attn_kwargs`` provided.

    Parameters
    ----------
    query : torch.Tensor
        Tensor of shape (..., items_out, channels)
    key : torch.Tensor
        Tensor of shape (..., items_in, channels)
    value : torch.Tensor
        Tensor of shape (..., items_in, channels)
    **attn_kwargs
        Optional keyword arguments passed to attention.

    Returns
    -------
    torch.Tensor
        Tensor of shape (..., head, item_out, channels)
    """
    attention_backend = get_attention_backend(**attn_kwargs)
    return attention_backend(query, key, value, **attn_kwargs)



import torch
import pytest
from tests.constants import TOLERANCES, BATCH_DIMS

from lloca.utils.orthogonalize_3d import orthogonalize_3d


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("method", ["cross", "gramschmidt"])
@pytest.mark.parametrize(
    "vector_type,eps",
    [("naive", None), ("collinear", 1e-2), ("collinear", 1e-3), ("collinear", 1e-4)],
)
def test_orthogonalize_o3(batch_dims, method, vector_type, eps):
    dtype = torch.float64

    v1 = torch.randn(batch_dims + [3], dtype=dtype)
    if vector_type == "naive":
        v2 = torch.randn_like(v1)
    elif vector_type == "collinear":
        v2 = v1 + eps * torch.randn_like(v1)
    else:
        raise ValueError(f"vector_type {vector_type} not implemented")

    orthogonal_vecs = orthogonalize_3d([v1, v2], method=method)

    # test orthonormality
    for i1, v1 in enumerate(orthogonal_vecs):
        for i2, v2 in enumerate(orthogonal_vecs):
            inner = (v1 * v2).sum(dim=-1)
            target = torch.ones_like(inner) if i1 == i2 else torch.zeros_like(inner)
            torch.testing.assert_close(inner, target, **TOLERANCES)



import torch
import pytest
from torch.nn import Linear
from tests.constants import TOLERANCES, LOGM2_MEAN_STD, REPS, FRAMES_PREDICTOR
from tests.helpers import sample_particle, equivectors_builder

from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.backbone.attention import LLoCaAttention
from lloca.framesnet.frames import InverseFrames
from lloca.utils.rand_transforms import rand_lorentz


@pytest.mark.parametrize("FramesPredictor", FRAMES_PREDICTOR)
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("hidden_reps", REPS)
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_invariance_equivariance(
    FramesPredictor,
    batch_dims,
    hidden_reps,
    logm2_std,
    logm2_mean,
):
    dtype = torch.float64

    # preparations
    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # preparations
    in_reps = TensorReps("1x1n")
    hidden_reps = TensorReps(hidden_reps)
    trafo = TensorRepsTransform(TensorReps(in_reps))
    attention = LLoCaAttention(hidden_reps, 1).to(dtype=dtype)
    linear_in = Linear(in_reps.dim, 3 * hidden_reps.dim).to(dtype=dtype)
    linear_out = Linear(hidden_reps.dim, in_reps.dim).to(dtype=dtype)

    # random global transformation
    random = rand_lorentz([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)

    # path 1: Frames transform + random transform
    frames = call_predictor(fm)
    attention.prepare_frames(frames)
    fm_local = trafo(fm, frames)
    x_local = linear_in(fm_local).unsqueeze(0)
    q_local, k_local, v_local = x_local.chunk(3, dim=-1)
    x_local2 = attention(q_local, k_local, v_local).squeeze(0)
    fm_local = linear_out(x_local2)
    fm_global = trafo(fm_local, InverseFrames(frames))
    fm_global_prime = torch.einsum("...ij,...j->...i", random, fm_global)

    # path 2: random transform + Frames transform
    fm_prime = torch.einsum("...ij,...j->...i", random, fm)
    frames_prime = call_predictor(fm_prime)
    attention.prepare_frames(frames_prime)
    fm_prime_local = trafo(fm_prime, frames_prime)
    x_prime_local = linear_in(fm_prime_local).unsqueeze(0)
    q_prime_local, k_prime_local, v_prime_local = x_prime_local.chunk(3, dim=-1)
    x_prime_local2 = attention(q_prime_local, k_prime_local, v_prime_local).squeeze(0)
    fm_prime_local = linear_out(x_prime_local2)
    fm_prime_global = trafo(fm_prime_local, InverseFrames(frames_prime))

    # test feature invariance before the operation
    torch.testing.assert_close(x_local, x_prime_local, **TOLERANCES)

    # test feature invariance after the operation
    torch.testing.assert_close(x_local2, x_prime_local2, **TOLERANCES)

    # test equivariance of output
    torch.testing.assert_close(fm_prime_global, fm_global_prime, **TOLERANCES)



import torch

from .lorentz import lorentz_squarednorm


def get_batch_from_ptr(ptr):
    """Reconstruct batch indices (batch) from pointer (ptr).

    Parameters
    ----------
    ptr : torch.Tensor
        Pointer tensor indicating the start of each batch.

    Returns
    -------
    torch.Tensor
        A tensor where each element indicates the batch index for each item.
    """
    return torch.arange(len(ptr) - 1, device=ptr.device).repeat_interleave(
        ptr[1:] - ptr[:-1],
    )


def get_ptr_from_batch(batch):
    """Reconstruct pointer (ptr) from batch indices (batch).

    Parameters
    ----------
    batch : torch.Tensor
        A tensor where each element indicates the batch index for each item.

    Returns
    -------
    torch.Tensor
        A pointer tensor indicating the start of each batch.
    """
    return torch.cat(
        [
            torch.tensor([0], device=batch.device),
            torch.where(batch[1:] - batch[:-1] != 0)[0] + 1,
            torch.tensor([batch.shape[0]], device=batch.device),
        ],
        0,
    )


def get_edge_index_from_ptr(ptr, remove_self_loops=True):
    """Construct edge index of fully connected graph from pointer (ptr).

    Parameters
    ----------
    ptr : torch.Tensor
        Pointer tensor indicating the start of each batch.
    remove_self_loops : bool, optional
        Whether to remove self-loops from the edge index, by default True.

    Returns
    -------
    torch.Tensor
        A tensor of shape (2, E) where E is the number of edges, representing the edge index.
    """
    row = torch.arange(ptr.max(), device=ptr.device)
    diff = ptr[1:] - ptr[:-1]
    repeats = (diff).repeat_interleave(diff)
    row = row.repeat_interleave(repeats)

    repeater = torch.stack(
        (-diff + 1, torch.ones_like(diff, device=ptr.device))
    ).T.reshape(-1)
    extras = repeater.repeat_interleave(repeater.abs())
    integ = torch.ones(row.shape[0], dtype=torch.long, device=ptr.device)
    mask = (row[1:] - row[:-1]).to(torch.bool)
    integ[0] = 0
    integ[1:][mask] = extras[:-1]
    col = torch.cumsum(integ, 0)

    edge_index = torch.stack((row, col))

    if remove_self_loops:
        row, col = edge_index
        edge_index = edge_index[:, row != col]

    return edge_index


def build_edge_index_fully_connected(features_ref, remove_self_loops=True):
    """Construct edge index of fully connected graph from reference object.
    Only shape and device of the reference object are used.

    Parameters
    ----------
    features_ref : torch.Tensor
        Reference tensor from which the shape and device are derived.
    remove_self_loops : bool, optional
        Whether to remove self-loops from the edge index, by default True.

    Returns
    -------
    torch.Tensor
        A tensor of shape (2, E) where E is the number of edges, representing the edge index.
    """
    B, N, _ = features_ref.shape
    device = features_ref.device

    nodes = torch.arange(N, device=device)
    row = nodes.repeat_interleave(N)
    col = nodes.repeat(N)

    if remove_self_loops:
        mask = row != col
        row, col = row[mask], col[mask]

    edge_base = torch.stack([row, col], dim=0)

    offsets = torch.arange(B, device=device, dtype=torch.long) * N
    batched = edge_base.unsqueeze(2) + offsets.view(1, 1, -1)
    edge_index_global = batched.permute(0, 2, 1).reshape(2, -1)

    batch = torch.arange(B, device=device).repeat_interleave(N)
    return edge_index_global, batch


def get_edge_attr(fourmomenta, edge_index, eps=1e-10, use_float64=True):
    """Calculate edge attributes based on the squared Lorentz norm of the sum of four-momenta.

    Parameters
    ----------
    fourmomenta : torch.Tensor
        A tensor of shape (B, N, 4) representing the four-momenta of particles.
    edge_index : torch.Tensor
        A tensor of shape (2, E) representing the edge index of the graph.
    eps : float, optional
        A small value to avoid log(0) issues, by default 1e-10.
    use_float64 : bool, optional
        Whether to use float64 precision for calculations, by default True.

    Returns
    -------
    torch.Tensor
        A tensor of shape (E,) representing the edge attributes, which are the logarithm of the squared Lorentz norm.
    """
    if use_float64:
        in_dtype = fourmomenta.dtype
        fourmomenta = fourmomenta.to(torch.float64)
    mij2 = lorentz_squarednorm(fourmomenta[edge_index[0]] + fourmomenta[edge_index[1]])
    edge_attr = mij2.clamp(min=eps).log()
    if use_float64:
        edge_attr = edge_attr.to(in_dtype)
    return edge_attr



"""Particle Transformer (ParT)

Paper: "Particle Transformer for Jet Tagging" - https://arxiv.org/abs/2202.03772

We have to do three things to build a LLoCa-ParT
- Construct a LLoCaAttention module for the whole transformer that preprocesses the frames
  and is passed to each attention block during initialization.
- When evaluating attention, use the LLoCaAttention module.

More comments:
- We also added an extra clamp in to_ptrapphim to avoid numerical issues from log(0). This case
  might not happen in the original ParT, but it can happen with LLoCa for highly boosted frames.
- We use LLoCaAttention only for the self-attention blocks, and use default attention (corresponds
  to scalar messages only) for the class attention blocks.

Use git diff --no-index experiments/baselines/particletransformer.py lloca/nn/particletransformer.py
to see the changes required to include frame-to-frame transformations
"""

import math
import random
import copy
from functools import partial
from typing import Optional, Tuple, Any, Callable

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..reps.tensorreps import TensorReps
from .attention import LLoCaAttention


@torch.jit.script
def delta_phi(a, b):
    return (a - b + math.pi) % (2 * math.pi) - math.pi


@torch.jit.script
def delta_r2(eta1, phi1, eta2, phi2):
    return (eta1 - eta2) ** 2 + delta_phi(phi1, phi2) ** 2


def to_pt2(x, eps=1e-8):
    pt2 = x[:, :2].square().sum(dim=1, keepdim=True)
    if eps is not None:
        pt2 = pt2.clamp(min=eps)
    return pt2


def to_m2(x, eps=1e-8):
    m2 = x[:, 3:4].square() - x[:, :3].square().sum(dim=1, keepdim=True)
    if eps is not None:
        m2 = m2.clamp(min=eps)
    return m2


def to_ptrapphim(x, return_mass=True, eps=1e-8):
    # x: (N, 4, ...), dim1 : (px, py, pz, E)
    px, py, pz, energy = x.split((1, 1, 1, 1), dim=1)
    pt = torch.sqrt(to_pt2(x, eps=eps))
    # rapidity = 0.5 * torch.log((energy + pz) / (energy - pz))
    rapidity = 0.5 * torch.log(
        (1 + (2 * pz) / (energy - pz).clamp(min=1e-20)).clamp(min=1e-20)
    )
    phi = torch.atan2(py, px)
    if not return_mass:
        return torch.cat((pt, rapidity, phi), dim=1)
    else:
        m = torch.sqrt(to_m2(x, eps=eps))
        return torch.cat((pt, rapidity, phi, m), dim=1)


def boost(x, boostp4, eps=1e-8):
    # boost x to the rest frame of boostp4
    # x: (N, 4, ...), dim1 : (px, py, pz, E)
    p3 = -boostp4[:, :3] / boostp4[:, 3:].clamp(min=eps)
    b2 = p3.square().sum(dim=1, keepdim=True)
    gamma = (1 - b2).clamp(min=eps) ** (-0.5)
    gamma2 = (gamma - 1) / b2
    gamma2.masked_fill_(b2 == 0, 0)
    bp = (x[:, :3] * p3).sum(dim=1, keepdim=True)
    v = x[:, :3] + gamma2 * bp * p3 + x[:, 3:] * gamma * p3
    return v


def p3_norm(p, eps=1e-8):
    return p[:, :3] / p[:, :3].norm(dim=1, keepdim=True).clamp(min=eps)


def to_energy_momentum(x, return_unit_vector=True):
    energy = x[:, 3:4]
    p2 = x[:, :3].square().sum(dim=1, keepdim=True)
    mom = torch.sqrt(p2)
    if return_unit_vector:
        return energy, mom, x[:, :3] / mom.clamp(min=1e-8)
    else:
        return energy, mom


def to_cos_sin_angles(xi, xj, normed_inputs=False, eps=1e-8):
    if normed_inputs:
        ni, nj = xi, xj
    else:
        ni, nj = p3_norm(xi, eps), p3_norm(xj, eps)
    cos = (ni * nj).sum(dim=1, keepdim=True).clamp(min=-1, max=1)
    sin = (
        torch.linalg.cross(ni, nj, dim=1).norm(dim=1, keepdim=True).clamp(min=0, max=1)
    )
    return cos, sin


def pairwise_lv_fts_pp(xi, xj, num_outputs=4, eps=1e-8):
    pti, rapi, phii = to_ptrapphim(xi, False, eps=None).split((1, 1, 1), dim=1)
    ptj, rapj, phij = to_ptrapphim(xj, False, eps=None).split((1, 1, 1), dim=1)

    # modified this for convenience (only lorentz scalars is most conservative)
    xij = xi + xj
    lnm2 = torch.log(to_m2(xij, eps=eps))
    if num_outputs > 0:
        outputs = [lnm2]

    if num_outputs > 1:
        delta = delta_r2(rapi, phii, rapj, phij).sqrt()
        lndelta = torch.log(delta.clamp(min=eps))
        ptmin = torch.minimum(pti, ptj)
        lnkt = torch.log((ptmin * delta).clamp(min=eps))
        lnz = torch.log((ptmin / (pti + ptj).clamp(min=eps)).clamp(min=eps))
        outputs += [lnkt, lnz, lndelta]

    if num_outputs > 4:
        lnds2 = torch.log(torch.clamp(-to_m2(xi - xj, eps=None), min=eps))
        outputs.append(lnds2)

    # the following features are not symmetric for (i, j)
    if num_outputs > 5:
        xj_boost = boost(xj, xij)
        costheta = (p3_norm(xj_boost, eps=eps) * p3_norm(xij, eps=eps)).sum(
            dim=1, keepdim=True
        )
        outputs.append(costheta)

    if num_outputs > 6:
        deltarap = rapi - rapj
        deltaphi = delta_phi(phii, phij)
        outputs += [deltarap, deltaphi]

    assert len(outputs) == num_outputs
    return torch.cat(outputs, dim=1)


def pairwise_lv_fts_ee(xi, xj, num_outputs=6, eps=1e-8):
    # outputs: [lnm2, cos_angle, sin_angle, lnkt, lnz, lnjade]
    lnm2 = torch.log(to_m2(xi + xj, eps=eps))
    outputs = [lnm2]

    if num_outputs > 1:
        ei, pi, ni = to_energy_momentum(xi)
        ej, pj, nj = to_energy_momentum(xj)
        cos_angle, sin_angle = to_cos_sin_angles(ni, nj, normed_inputs=True)
        outputs += [cos_angle, sin_angle]

    if num_outputs > 3:
        pmin = torch.minimum(pi, pj)
        lnkt = torch.log((pmin * sin_angle).clamp(min=eps))
        lnz = torch.log((pmin / (pi + pj).clamp(min=eps)).clamp(min=eps))
        outputs += [lnkt, lnz]

    if num_outputs > 5:
        lnjade = torch.log((ei * ej * (1 - cos_angle)).clamp(min=eps))
        outputs.append(lnjade)

    assert len(outputs) == num_outputs
    return torch.cat(outputs, dim=1)


def build_sparse_tensor(uu, idx, seq_len):
    # inputs: uu (N, C, num_pairs), idx (N, 2, num_pairs)
    # return: (N, C, seq_len, seq_len)
    batch_size, num_fts, num_pairs = uu.size()
    idx = torch.min(idx, torch.ones_like(idx) * seq_len)
    i = torch.cat(
        (
            torch.arange(0, batch_size, device=uu.device)
            .repeat_interleave(num_fts * num_pairs)
            .unsqueeze(0),
            torch.arange(0, num_fts, device=uu.device)
            .repeat_interleave(num_pairs)
            .repeat(batch_size)
            .unsqueeze(0),
            idx[:, :1, :].expand_as(uu).flatten().unsqueeze(0),
            idx[:, 1:, :].expand_as(uu).flatten().unsqueeze(0),
        ),
        dim=0,
    )
    return torch.sparse_coo_tensor(
        i,
        uu.flatten(),
        size=(batch_size, num_fts, seq_len + 1, seq_len + 1),
        device=uu.device,
    ).to_dense()[:, :, :seq_len, :seq_len]


def tril_indices(row, col, offset=0, *, dtype=torch.long, device="cpu"):
    return torch.ones(row, col, dtype=dtype, device=device).tril(offset).nonzero().T


class SequenceTrimmer(nn.Module):
    def __init__(
        self, enabled=False, target=(0.9, 1.02), warmup_steps=5, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.enabled = enabled
        self.target = target
        self.warmup_steps = warmup_steps
        self.register_buffer("_counter", torch.LongTensor([0]), persistent=False)

    def forward(self, x, v=None, mask=None, uu=None):
        # x: (N, C, P)
        # v: (N, 4, P) [px,py,pz,energy]
        # mask: (N, 1, P) -- real particle = 1, padded = 0
        # uu: (N, C', P, P)
        if mask is None:
            mask = torch.ones_like(x[:, :1])
        mask = mask.bool()

        if self.enabled:
            if self._counter < self.warmup_steps:
                self._counter.add_(1)
            else:
                if v is not None:
                    if not isinstance(v, (list, tuple)):
                        v = [v]
                if self.training:
                    q = min(1, random.uniform(*self.target))
                    maxlen = torch.quantile(mask.float().sum(dim=-1), q).long()
                    rand = torch.rand_like(mask.float())
                    rand.masked_fill_(~mask, -1)
                    perm = rand.argsort(dim=-1, descending=True)  # (N, 1, P)
                    mask = torch.gather(mask, -1, perm)
                    x = torch.gather(x, -1, perm.expand_as(x))
                    if v is not None:
                        v = [torch.gather(_v, -1, perm.expand_as(_v)) for _v in v]
                    if uu is not None:
                        uu = torch.gather(uu, -2, perm.unsqueeze(-1).expand_as(uu))
                        uu = torch.gather(uu, -1, perm.unsqueeze(-2).expand_as(uu))
                else:
                    maxlen = mask.sum(dim=-1).max()
                maxlen = max(maxlen, 1)
                if maxlen < mask.size(-1):
                    mask = mask[:, :, :maxlen]
                    x = x[:, :, :maxlen]
                    if v is not None:
                        v = [_v[:, :, :maxlen] for _v in v]
                    if uu is not None:
                        uu = uu[:, :, :maxlen, :maxlen]
                if v is not None:
                    if len(v) == 1:
                        v = v[0]

        return x, v, mask, uu


class SwiGLUFFN(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: Optional[int] = None,
        out_features: Optional[int] = None,
        drop: float = 0.0,
        bias: bool = True,
    ) -> None:
        super().__init__()
        hidden_features = hidden_features or in_features
        out_features = out_features or in_features
        self.w12 = nn.Linear(in_features, 2 * hidden_features, bias=bias)
        self.drop = nn.Dropout(drop)
        self.w3 = nn.Linear(hidden_features, out_features, bias=bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x12 = self.w12(x)
        x1, x2 = x12.chunk(2, dim=-1)
        hidden = F.silu(x1) * x2
        hidden = self.drop(hidden)
        return self.w3(hidden)


class Embed(nn.Module):
    def __init__(self, input_dim, dims, normalize_input=True, activation="gelu"):
        super().__init__()

        self.input_bn = nn.BatchNorm1d(input_dim) if normalize_input else None
        module_list = []
        for dim in dims:
            module_list.extend(
                [
                    nn.LayerNorm(input_dim),
                    nn.Linear(input_dim, dim),
                    nn.GELU() if activation == "gelu" else nn.ReLU(),
                ]
            )
            input_dim = dim
        self.embed = nn.Sequential(*module_list)

    def forward(self, x):
        if self.input_bn is not None:
            # x: (batch, embed_dim, seq_len)
            x = self.input_bn(x)
            x = x.transpose(1, 2).contiguous()
        # x: (batch, seq_len, embed_dim)
        return self.embed(x)


class PairEmbed(nn.Module):
    def __init__(
        self,
        pairwise_lv_dim,
        pairwise_input_dim,
        dims,
        pairwise_lv_type="pp",
        remove_self_pair=False,
        use_pre_activation_pair=True,
        normalize_input=True,
        activation="gelu",
        eps=1e-8,
        for_onnx=False,
        sparse_eval=None,
    ):
        super().__init__()

        self.pairwise_lv_dim = pairwise_lv_dim
        self.pairwise_input_dim = pairwise_input_dim
        self.remove_self_pair = remove_self_pair
        self.for_onnx = for_onnx
        self.sparse_eval = (not for_onnx) if sparse_eval is None else sparse_eval
        self.out_dim = dims[-1]

        if pairwise_lv_type == "pp":
            self.is_symmetric = (pairwise_lv_dim <= 5) and (pairwise_input_dim == 0)
            self.pairwise_lv_fts = partial(
                pairwise_lv_fts_pp, num_outputs=pairwise_lv_dim, eps=eps
            )
        elif pairwise_lv_type == "ee":
            self.is_symmetric = (pairwise_lv_dim <= 6) and (pairwise_input_dim == 0)
            self.pairwise_lv_fts = partial(
                pairwise_lv_fts_ee, num_outputs=pairwise_lv_dim, eps=eps
            )
        else:
            raise RuntimeError(
                "Invalid value for `pairwise_lv_type`: " + pairwise_lv_type
            )

        if pairwise_lv_dim > 0:
            input_dim = pairwise_lv_dim
            module_list = [nn.BatchNorm1d(input_dim)] if normalize_input else []
            for dim in dims:
                module_list.extend(
                    [
                        nn.Conv1d(input_dim, dim, 1),
                        nn.BatchNorm1d(dim),
                        nn.GELU() if activation == "gelu" else nn.ReLU(),
                    ]
                )
                input_dim = dim
            if use_pre_activation_pair:
                module_list = module_list[:-1]
            self.embed = nn.Sequential(*module_list)

        if pairwise_input_dim > 0:
            input_dim = pairwise_input_dim
            module_list = [nn.BatchNorm1d(input_dim)] if normalize_input else []
            for dim in dims:
                module_list.extend(
                    [
                        nn.Conv1d(input_dim, dim, 1),
                        nn.BatchNorm1d(dim),
                        nn.GELU() if activation == "gelu" else nn.ReLU(),
                    ]
                )
                input_dim = dim
            if use_pre_activation_pair:
                module_list = module_list[:-1]
            self.fts_embed = nn.Sequential(*module_list)

    def _forward_dense(self, x, uu=None, mask=None):
        # x: (batch, v_dim, seq_len)
        # uu: (batch, v_dim, seq_len, seq_len)
        assert x is not None or uu is not None
        with torch.no_grad():
            if x is not None:
                batch_size, _, seq_len = x.size()
            else:
                batch_size, _, seq_len, _ = uu.size()
            if self.is_symmetric:
                tril_indices_fn = tril_indices if self.for_onnx else torch.tril_indices
                i, j = tril_indices_fn(
                    seq_len,
                    seq_len,
                    offset=-1 if self.remove_self_pair else 0,
                    device=(x if x is not None else uu).device,
                )
                if x is not None:
                    x = x.unsqueeze(-1).repeat(1, 1, 1, seq_len)
                    xi = x[:, :, i, j]  # (batch, dim, seq_len*(seq_len+1)/2)
                    xj = x[:, :, j, i]
                    x = self.pairwise_lv_fts(xi, xj)
                if uu is not None:
                    # (batch, dim, seq_len*(seq_len+1)/2)
                    uu = uu[:, :, i, j]
            else:
                if x is not None:
                    x = self.pairwise_lv_fts(x.unsqueeze(-1), x.unsqueeze(-2))
                    if self.remove_self_pair:
                        i = torch.arange(0, seq_len, device=x.device)
                        x[:, :, i, i] = 0
                    x = x.view(-1, self.pairwise_lv_dim, seq_len * seq_len)
                if uu is not None:
                    uu = uu.view(-1, self.pairwise_input_dim, seq_len * seq_len)

        # with grad
        elements = 0
        if x is not None:
            elements = elements + self.embed(x)
        if uu is not None:
            elements = elements + self.fts_embed(uu)

        if self.is_symmetric:
            y = torch.zeros(
                batch_size,
                self.out_dim,
                seq_len,
                seq_len,
                dtype=elements.dtype,
                device=elements.device,
            )
            y[:, :, i, j] = elements
            y[:, :, j, i] = elements
        else:
            y = elements.view(-1, self.out_dim, seq_len, seq_len)
        return y

    def _forward_sparse(self, x, uu=None, mask=None):
        # x: (batch, v_dim, seq_len)
        # uu: (batch, v_dim, seq_len, seq_len)
        assert x is not None or uu is not None
        with torch.no_grad():
            if x is not None:
                batch_size, _, seq_len = x.size()
            else:
                batch_size, _, seq_len, _ = uu.size()

            i0, i1, i2, i3 = (Ellipsis,) * 4
            if mask is not None:
                mask = mask.unsqueeze(-1) * mask.unsqueeze(
                    -2
                )  # (batch_size, 1, seq_len, seq_len)
                if self.is_symmetric:
                    offset = -1 if self.remove_self_pair else 0
                    i0, _, i2, i3 = mask.float().tril(offset).nonzero(as_tuple=True)
                else:
                    i0, _, i2, i3 = mask.nonzero(as_tuple=True)

            if x is not None:
                x = self.pairwise_lv_fts(x.unsqueeze(-1), x.unsqueeze(-2))
                x = x.permute(0, 2, 3, 1)[
                    i0, i2, i3, :
                ]  # (num_elements, pairwise_lv_dim)
                x = x.T.unsqueeze(0).contiguous()  # (1, pairwise_lv_dim, num_elements)
            if uu is not None:
                uu = uu.permute(0, 2, 3, 1)[
                    i0, i2, i3, :
                ]  # (num_elements, pairwise_input_dim)
                uu = uu.T.unsqueeze(
                    0
                ).contiguous()  # (1, pairwise_input_dim, num_elements)

        # with grad
        elements = 0
        if x is not None:
            elements = elements + self.embed(x)
        if uu is not None:
            elements = elements + self.fts_embed(uu)
        elements = elements.squeeze(0).T  # (num_elements, out_dim)

        y = torch.zeros(
            batch_size,
            seq_len,
            seq_len,
            self.out_dim,
            dtype=elements.dtype,
            device=elements.device,
        )
        y[i0, i2, i3, :] = elements
        if self.is_symmetric:
            y[i0, i3, i2, :] = elements
        y = y.permute(0, 3, 1, 2).contiguous()

        return y

    def forward(self, x, uu=None, mask=None):
        if self.sparse_eval:
            return self._forward_sparse(x, uu=uu, mask=mask)
        else:
            return self._forward_dense(x, uu=uu, mask=mask)


def _canonical_mask(
    mask: Optional[torch.Tensor],
    mask_name: str,
    other_type: Optional[Any],
    other_name: str,
    target_type: Any,
    check_other: bool = True,
) -> Optional[torch.Tensor]:

    if mask is not None:
        _mask_dtype = mask.dtype
        _mask_is_float = torch.is_floating_point(mask)
        if _mask_dtype != torch.bool and not _mask_is_float:
            raise AssertionError(
                f"only bool and floating types of {mask_name} are supported"
            )
        if not _mask_is_float:
            mask = torch.zeros_like(mask, dtype=target_type).masked_fill_(
                mask, float("-inf")
            )
    return mask


def _none_or_dtype(input: Optional[torch.Tensor]):
    if input is None:
        return None
    elif isinstance(input, torch.Tensor):
        return input.dtype
    raise RuntimeError("input to _none_or_dtype() must be None or torch.Tensor")


class Attention(torch.nn.Module):
    def __init__(
        self,
        attention,
        embed_dim,
        num_heads,
        dropout=0.0,
        bias=True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"

        self.in_proj = torch.nn.Linear(
            embed_dim, 3 * embed_dim, bias=bias, **factory_kwargs
        )
        self.out_proj = torch.nn.Linear(
            embed_dim, embed_dim, bias=bias, **factory_kwargs
        )
        self.attention = attention

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):

        for k in state_dict.keys():
            if k.endswith("in_proj_weight"):
                state_dict[k.replace("_weight", ".weight")] = state_dict.pop(k)
            elif k.endswith("in_proj_bias"):
                state_dict[k.replace("_bias", ".bias")] = state_dict.pop(k)

        super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
        attn_mask: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:

        bsz, tgt_len, _ = query.shape
        _, src_len, _ = key.shape

        # (bsz, src_len)
        key_padding_mask = _canonical_mask(
            mask=key_padding_mask,
            mask_name="key_padding_mask",
            other_type=_none_or_dtype(attn_mask),
            other_name="attn_mask",
            target_type=query.dtype,
        )

        # (bsz, num_heads, tgt_len, src_len)
        attn_mask = _canonical_mask(
            mask=attn_mask,
            mask_name="attn_mask",
            other_type=None,
            other_name="",
            target_type=query.dtype,
            check_other=False,
        )

        # merge key padding and attention masks
        if key_padding_mask is not None:
            assert key_padding_mask.shape == (
                bsz,
                src_len,
            ), f"expecting key_padding_mask shape of {(bsz, src_len)}, but got {key_padding_mask.shape}"
            key_padding_mask = key_padding_mask.view(bsz, 1, 1, src_len).expand(
                -1, self.num_heads, -1, -1
            )
            if attn_mask is None:
                attn_mask = key_padding_mask
            else:
                assert attn_mask.shape == (
                    bsz,
                    self.num_heads,
                    tgt_len,
                    src_len,
                ), f"expecting attn_mask shape of {(bsz, self.num_heads, tgt_len, src_len)}, but got {attn_mask.shape}"
                attn_mask = attn_mask + key_padding_mask

        # (bsz, seq_len, num_heads*head_dim)
        q, k, v = F._in_projection_packed(
            query, key, value, self.in_proj.weight, self.in_proj.bias
        )

        # -> (bsz, num_heads, src/tgt_len, head_dim)
        q = (
            q.view(bsz, tgt_len, self.num_heads, self.head_dim)
            .transpose(1, 2)
            .contiguous()
        )
        k = (
            k.view(bsz, src_len, self.num_heads, self.head_dim)
            .transpose(1, 2)
            .contiguous()
        )
        v = (
            v.view(bsz, src_len, self.num_heads, self.head_dim)
            .transpose(1, 2)
            .contiguous()
        )

        dropout_p = self.dropout if self.training else 0.0

        if self.attention is not None:
            # particle attention
            attn_output = self.attention(
                q,
                k,
                v,
                attn_mask=attn_mask,
                dropout_p=dropout_p,
            )
        else:
            # class token attention
            attn_output = F.scaled_dot_product_attention(q, k, v, attn_mask, dropout_p)

        attn_output = attn_output.transpose(1, 2).reshape(bsz, tgt_len, self.embed_dim)
        attn_output = self.out_proj(attn_output)
        return attn_output, None


class LayerScale(nn.Module):
    def __init__(
        self,
        dim: int,
        init_values: float = 1e-5,
        inplace: bool = False,
    ) -> None:
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.mul_(self.gamma) if self.inplace else x * self.gamma


def drop_path(
    x, drop_prob: float = 0.0, training: bool = False, scale_by_keep: bool = True
):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob,3):0.3f}"


class Block(nn.Module):
    def __init__(
        self,
        attention,
        embed_dim=128,
        num_heads=8,
        ffn_ratio=4,
        dropout=0.1,
        attn_dropout=0.1,
        activation_dropout=0.1,
        activation="gelu",
        layer_scale_init_values=None,
        drop_path_rate=0.0,
        scale_attn_mask=False,
        scale_attn=True,
        scale_fc=True,
        scale_heads=True,
        scale_resids=True,
    ):
        super().__init__()

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.ffn_dim = embed_dim * ffn_ratio

        self.pre_attn_norm = nn.LayerNorm(embed_dim)
        self.attn = Attention(attention, embed_dim, num_heads, dropout=attn_dropout)
        self.post_attn_norm = nn.LayerNorm(embed_dim) if scale_attn else nn.Identity()
        self.dropout = nn.Dropout(dropout)
        self.ls1 = (
            LayerScale(embed_dim, init_values=layer_scale_init_values)
            if layer_scale_init_values
            else nn.Identity()
        )
        self.drop_path1 = (
            DropPath(drop_path_rate) if drop_path_rate > 0.0 else nn.Identity()
        )

        self.pre_fc_norm = nn.LayerNorm(embed_dim)
        self.fc1 = nn.Linear(embed_dim, self.ffn_dim)
        if activation == "swiglu":
            self.fc1_g = nn.Linear(embed_dim, self.ffn_dim)
            self.act = nn.SiLU()
        else:
            self.fc1_g = None
            self.act = nn.GELU() if activation == "gelu" else nn.ReLU()
        self.act_dropout = nn.Dropout(activation_dropout)
        self.post_fc_norm = nn.LayerNorm(self.ffn_dim) if scale_fc else nn.Identity()
        self.fc2 = nn.Linear(self.ffn_dim, embed_dim)
        self.ls2 = (
            LayerScale(embed_dim, init_values=layer_scale_init_values)
            if layer_scale_init_values
            else nn.Identity()
        )
        self.drop_path2 = (
            DropPath(drop_path_rate) if drop_path_rate > 0.0 else nn.Identity()
        )

        self.c_mask = (
            nn.Parameter(torch.ones(1), requires_grad=True) if scale_attn_mask else None
        )
        self.c_attn = (
            nn.Parameter(torch.ones(num_heads), requires_grad=True)
            if scale_heads
            else None
        )
        self.w_resid = (
            nn.Parameter(torch.ones(embed_dim), requires_grad=True)
            if scale_resids
            else None
        )

    def forward(self, x, x_cls=None, padding_mask=None, attn_mask=None):
        """
        Args:
            x (Tensor): input to the layer of shape `(batch, seq_len, embed_dim)`
            x_cls (Tensor, optional): class token input to the layer of shape `(batch, 1, embed_dim)`
            padding_mask (ByteTensor, optional): binary
                ByteTensor of shape `(batch, seq_len)` where padding
                elements are indicated by ``True``.

        Returns:
            encoded output of shape `(batch, seq_len, embed_dim)`
        """

        if x_cls is not None:
            with torch.no_grad():
                # prepend one element for x_cls: -> (batch, 1+seq_len)
                padding_mask = torch.cat(
                    (torch.zeros_like(padding_mask[:, :1]), padding_mask), dim=1
                )
            # class attention: https://arxiv.org/pdf/2103.17239.pdf
            residual = x_cls
            u = torch.cat((x_cls, x), dim=1)  # (batch, 1+seq_len, embed_dim)
            u = self.pre_attn_norm(u)

            # default attention for convenience (could be more fancy here)
            x = self.attn(x_cls, u, u, key_padding_mask=padding_mask,)[
                0
            ]  # (1, batch, embed_dim)
        else:
            if self.c_mask is not None and attn_mask is not None:
                attn_mask = torch.mul(self.c_mask, attn_mask)
            residual = x
            x = self.pre_attn_norm(x)
            x = self.attn(x, x, x, key_padding_mask=padding_mask, attn_mask=attn_mask)[
                0
            ]  # (batch, seq_len, embed_dim)

        if self.c_attn is not None:
            bsz, tgt_len, _ = x.size()
            x = x.view(bsz, tgt_len, self.num_heads, self.head_dim)
            x = torch.einsum("bthd,h->btdh", x, self.c_attn)
            x = x.reshape(bsz, tgt_len, self.embed_dim)
        x = self.post_attn_norm(x)
        x = self.dropout(x)
        x = self.drop_path1(self.ls1(x))
        x += residual

        residual = x
        x = self.pre_fc_norm(x)
        if self.fc1_g is None:
            x = self.act(self.fc1(x))
        else:
            x_gate = self.fc1_g(x)
            x = self.fc1(x)
            x = self.act(x_gate) * x
        x = self.act_dropout(x)
        x = self.post_fc_norm(x)
        x = self.fc2(x)
        x = self.dropout(x)
        x = self.drop_path2(self.ls2(x))
        if self.w_resid is not None:
            residual = torch.mul(self.w_resid, residual)
        x += residual

        return x


class ParticleTransformer(nn.Module):
    def __init__(
        self,
        input_dim,
        attn_reps,
        num_classes=None,
        # network configurations
        pair_input_type="pp",
        pair_input_dim=None,
        pair_extra_dim=0,
        remove_self_pair=False,
        use_pre_activation_pair=True,
        embed_dims=(128, 512, 128),
        pair_embed_dims=(64, 64, 64),
        num_heads=8,
        num_layers=8,
        num_cls_layers=2,
        block_params=None,
        cls_block_params=None,
        fc_params=(),
        activation="gelu",
        # misc
        version=1,
        weight_init="moco",
        fix_init=True,
        trim=True,
        for_inference=False,
        for_segmentation=False,
        use_amp=False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.trimmer = SequenceTrimmer(enabled=trim and not for_inference)
        self.for_inference = for_inference
        self.for_segmentation = for_segmentation
        self.use_amp = use_amp

        self.embed_dim = embed_dims[-1] if len(embed_dims) > 0 else input_dim
        attn_reps = TensorReps(attn_reps)
        assert attn_reps.dim * num_heads == self.embed_dim
        self.attention = LLoCaAttention(attn_reps, num_heads)
        default_cfg = dict(
            embed_dim=self.embed_dim,
            num_heads=num_heads,
            ffn_ratio=4,
            dropout=0.1,
            attn_dropout=0.1,
            activation_dropout=0.1,
            activation=activation,
            layer_scale_init_values=None,
            drop_path_rate=0.0,
            scale_attn_mask=False,
            scale_fc=True,
            scale_attn=True,
            scale_heads=True,
            scale_resids=True,
        )
        if version > 1:
            default_cfg.update(
                activation="swiglu",
                scale_fc=False,
                scale_attn=False,
                scale_heads=False,
                scale_resids=False,
            )

        cfg_block = copy.deepcopy(default_cfg)
        if block_params is not None:
            cfg_block.update(block_params)

        cfg_cls_block = copy.deepcopy(default_cfg)
        cfg_cls_block.update({"dropout": 0, "attn_dropout": 0, "activation_dropout": 0})
        if cls_block_params is not None:
            cfg_cls_block.update(cls_block_params)

        self.embed = (
            Embed(input_dim, embed_dims, activation=activation)
            if len(embed_dims) > 0
            else nn.Identity()
        )

        if pair_input_dim is None:
            pair_input_dim = 4 if pair_input_type == "pp" else 6
        self.pair_extra_dim = pair_extra_dim
        self.pair_embed = (
            PairEmbed(
                pair_input_dim,
                pair_extra_dim,
                (*pair_embed_dims, cfg_block["num_heads"]),
                pairwise_lv_type=pair_input_type,
                remove_self_pair=remove_self_pair,
                use_pre_activation_pair=use_pre_activation_pair,
                for_onnx=for_inference,
            )
            if pair_embed_dims is not None and pair_input_dim + pair_extra_dim > 0
            else None
        )
        self.blocks = nn.ModuleList(
            [Block(attention=self.attention, **cfg_block) for _ in range(num_layers)]
        )
        self.cls_blocks = (
            nn.ModuleList(
                [Block(attention=None, **cfg_cls_block) for _ in range(num_cls_layers)]
            )
            if num_cls_layers > 0
            else None
        )
        self.norm = nn.LayerNorm(self.embed_dim)

        if fc_params is not None:
            fcs = []
            in_dim = self.embed_dim
            for param in fc_params:
                try:
                    out_dim, drop_rate, act = param
                except ValueError:
                    (out_dim, drop_rate), act = param, "relu"
                if act == "swiglu":
                    layer = nn.Sequential(
                        SwiGLUFFN(in_dim, out_dim * 4, out_dim, drop=drop_rate),
                        nn.LayerNorm(out_dim),
                    )
                else:
                    layer = nn.Sequential(
                        nn.Linear(in_dim, out_dim),
                        nn.GELU() if act == "gelu" else nn.ReLU(),
                        nn.Dropout(drop_rate),
                    )
                fcs.append(layer)
                in_dim = out_dim
            fcs.append(nn.Linear(in_dim, num_classes))
            self.fc = nn.Sequential(*fcs)
        else:
            self.fc = None

        # cls tokens
        if not self.for_segmentation and num_cls_layers > 0:
            self.cls_token = nn.Parameter(
                torch.zeros(1, 1, self.embed_dim), requires_grad=True
            )
            nn.init.trunc_normal_(self.cls_token, std=0.02)
        else:
            self.cls_token = None

        # weight initialization
        if weight_init is not None:
            self.init_weights(weight_init)
        if fix_init:
            self.fix_init_weight()

    def fix_init_weight(self):
        def rescale(param, _layer_id):
            param.div_(math.sqrt(2.0 * _layer_id))

        for layer_id, layer in enumerate(self.blocks):
            rescale(layer.attn.out_proj.weight.data, layer_id + 1)
            rescale(layer.fc2.weight.data, layer_id + 1)

    def init_weights(self, mode: str = "") -> None:
        assert mode in ("timm", "moco")
        if mode == "timm":
            named_apply(init_weights_vit_timm, self)
        elif mode == "moco":
            named_apply(init_weights_vit_moco, self)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {
            "cls_token",
        }

    def _forward_encoder(self, x, v=None, mask=None, uu=None, uu_idx=None):
        with torch.no_grad():
            if not self.for_inference:
                if uu_idx is not None:
                    uu = build_sparse_tensor(uu, uu_idx, x.size(-1))
            x, v, mask, uu = self.trimmer(x, v, mask, uu)
            padding_mask = ~mask.squeeze(1)  # (batch_size, seq_len)

        with torch.autocast("cuda", enabled=self.use_amp):
            # input embedding
            x = self.embed(x).masked_fill(
                ~mask.transpose(1, 2), 0
            )  # (batch_size, seq_len, num_fts)
            attn_mask = None
            if (v is not None or uu is not None) and self.pair_embed is not None:
                attn_mask = self.pair_embed(
                    v, uu=uu, mask=mask
                )  # (batch_size, num_heads, seq_len, seq_len)

            # transform
            for block in self.blocks:
                x = block(x, x_cls=None, padding_mask=padding_mask, attn_mask=attn_mask)

        # x: (batch, seq_len, embed_dim)
        # padding_mask: (batch, seq_len)
        return x, padding_mask

    def _forward_aggregator(self, x, padding_mask):
        with torch.autocast("cuda", enabled=self.use_amp):
            if self.cls_blocks is not None:
                # for classification: extract using class token
                cls_tokens = self.cls_token.expand(
                    x.size(0), 1, -1
                )  # (batch, 1, embed_dim)
                for block in self.cls_blocks:
                    cls_tokens = block(
                        x, x_cls=cls_tokens, padding_mask=padding_mask
                    )  # (batch, 1, embed_dim)
                cls_tokens = cls_tokens.squeeze(1)  # (batch, embed_dim)
            else:
                # for classification: simple average pooling
                mask = ~padding_mask.unsqueeze(1)  # (batch, 1, seq_len)
                x = x.transpose(1, 2).contiguous()  # (batch, embed_dim, seq_len)
                counts = mask.float().sum(-1)  # (batch, 1)
                counts = torch.max(counts, torch.ones_like(counts))  # >=1
                cls_tokens = (x * mask).sum(-1) / counts  # (batch, embed_dim)

            x_cls = self.norm(cls_tokens)  # (batch, embed_dim)
        return x_cls

    def forward(self, x, frames, v=None, mask=None, uu=None, uu_idx=None):
        # x: (batch_size, num_fts, seq_len)
        # v: (batch_size, 4, seq_len) [px,py,pz,energy]
        # mask: (batch_size, 1, seq_len) -- real particle = 1, padded = 0
        # for pytorch: uu (batch_size, C', num_pairs), uu_idx (batch_size, 2, num_pairs)
        # for onnx: uu (batch_size, C', seq_len, seq_len), uu_idx=None
        self.attention.prepare_frames(frames)

        x, padding_mask = self._forward_encoder(x, v=v, mask=mask, uu=uu, uu_idx=uu_idx)

        if self.cls_blocks is None and self.fc is None:
            # x: (batch, seq_len, embed_dim)
            # padding_mask: (batch, seq_len)
            return x, padding_mask

        with torch.autocast("cuda", enabled=self.use_amp):
            # === for segmentation ===
            if self.for_segmentation:
                x = self.norm(x)
                if self.fc is not None:
                    x = self.fc(x)
                # x: (P, N, C) -> output: (N, C, P)
                output = x.transpose(1, 2).contiguous()
                if self.for_inference:
                    output = torch.softmax(output, dim=1)
                # print('output:\n', output)
                return output

            x_cls = self._forward_aggregator(x, padding_mask)
            if self.fc is None:
                return x_cls

            # fc
            output = self.fc(x_cls)
            if self.for_inference:
                output = torch.softmax(output, dim=1)
            # print('output:\n', output)
            return output


### weight initialization methods ###
def init_weights_vit_timm(module: nn.Module, name: str = "") -> None:
    """ViT weight initialization, original timm impl (for reproducibility)"""
    if isinstance(module, nn.Linear):
        nn.init.trunc_normal_(module.weight, std=0.02)
        if module.bias is not None:
            nn.init.zeros_(module.bias)
    elif hasattr(module, "init_weights"):
        module.init_weights()


def init_weights_vit_moco(module: nn.Module, name: str = "") -> None:
    """ViT weight initialization, matching moco-v3 impl minus fixed PatchEmbed"""
    if isinstance(module, nn.Linear):
        if "in_proj" in name:
            # treat the weights of Q, K, V separately
            val = math.sqrt(
                6.0 / float(module.weight.shape[0] // 3 + module.weight.shape[1])
            )
            nn.init.uniform_(module.weight, -val, val)
        else:
            nn.init.xavier_uniform_(module.weight)
        if module.bias is not None:
            nn.init.zeros_(module.bias)
    elif hasattr(module, "init_weights"):
        module.init_weights()


def named_apply(
    fn: Callable,
    module: nn.Module,
    name="",
    depth_first: bool = True,
    include_root: bool = False,
) -> nn.Module:
    if not depth_first and include_root:
        fn(module=module, name=name)
    for child_name, child_module in module.named_children():
        child_name = ".".join((name, child_name)) if name else child_name
        named_apply(
            fn=fn,
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
    if depth_first and include_root:
        fn(module=module, name=name)
    return module



"""Settings used for multiple tests."""

# Default tolerances
TOLERANCES = dict(atol=1e-3, rtol=1e-4)
MILD_TOLERANCES = dict(atol=0.05, rtol=0.05)
STRICT_TOLERANCES = dict(atol=1e-6, rtol=1e-6)

BATCH_DIMS = [[10, 10], [1000]]

REPS = ["4x0n", "4x1n", "10x0n+5x1n+2x2n"]

LOGM2_MEAN_STD = ((0, 1), (0, 0.1), (-3, 1))

from lloca.framesnet.equi_frames import (
    LearnedSO13Frames,
    LearnedRestFrames,
    LearnedPDFrames,
)

FRAMES_PREDICTOR = [
    LearnedSO13Frames,
    LearnedRestFrames,
    LearnedPDFrames,
]



import torch
import pytest
from tests.constants import TOLERANCES, LOGM2_MEAN_STD
from tests.helpers import sample_particle, lorentz_test, equivectors_builder

from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.utils.rand_transforms import rand_lorentz, rand_rotation, rand_xyrotation
from lloca.framesnet.frames import Frames
from lloca.framesnet.equi_frames import (
    LearnedSO13Frames,
    LearnedRestFrames,
    LearnedPDFrames,
    LearnedSO3Frames,
    LearnedSO2Frames,
)


@pytest.mark.parametrize(
    "FramesPredictor,rand_trafo",
    [
        (LearnedSO13Frames, rand_lorentz),
        (LearnedRestFrames, rand_lorentz),
        (LearnedPDFrames, rand_lorentz),
        (LearnedSO3Frames, rand_rotation),
        (LearnedSO2Frames, rand_xyrotation),
    ],
)
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_frames_transformation(
    FramesPredictor, rand_trafo, batch_dims, logm2_std, logm2_mean
):
    dtype = torch.float64

    # preparations
    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)

    # frames for un-transformed fm
    frames = call_predictor(fm)
    lorentz_test(frames.matrices, **TOLERANCES)

    # random global transformation
    random = rand_trafo([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # frames for transformed fm
    fm_prime = torch.einsum("...ij,...j->...i", random, fm)
    frames_prime = call_predictor(fm_prime)
    lorentz_test(frames_prime.matrices, **TOLERANCES)

    # check that frames transform correctly
    # expect frames_prime = frames * random^-1
    inv_random = Frames(random).inv
    frames_prime_expected = torch.einsum(
        "...ij,...jk->...ik", frames.matrices, inv_random
    )
    torch.testing.assert_close(
        frames_prime_expected, frames_prime.matrices, **TOLERANCES
    )


@pytest.mark.parametrize(
    "FramesPredictor,rand_trafo",
    [
        (LearnedSO13Frames, rand_lorentz),
        (LearnedRestFrames, rand_lorentz),
        (LearnedPDFrames, rand_lorentz),
        (LearnedSO3Frames, rand_rotation),
        (LearnedSO2Frames, rand_xyrotation),
    ],
)
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_feature_invariance(
    FramesPredictor, rand_trafo, batch_dims, logm2_std, logm2_mean
):
    dtype = torch.float64

    # preparations
    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    reps = TensorReps("1x1n")
    trafo = TensorRepsTransform(TensorReps(reps))

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)

    # random global transformation
    random = rand_trafo([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # path 1: Frames transform (+ random transform)
    frames = call_predictor(fm)
    lorentz_test(frames.matrices, **TOLERANCES)
    fm_local = trafo(fm, frames)

    # path 2: random transform + Frames transform
    fm_prime = torch.einsum("...ij,...j->...i", random, fm)
    frames_prime = call_predictor(fm_prime)
    lorentz_test(frames_prime.matrices, **TOLERANCES)
    fm_local_prime = trafo(fm_prime, frames_prime)

    # test that features are invariant
    torch.testing.assert_close(fm_local, fm_local_prime, **TOLERANCES)



import torch
import pytest
from tests.constants import TOLERANCES, LOGM2_MEAN_STD
from tests.helpers import sample_particle

from lloca.equivectors.equimlp import EquiMLP
from lloca.utils.rand_transforms import rand_lorentz


@pytest.mark.parametrize("batch_dims", [[100]])
@pytest.mark.parametrize("jet_size", [10])
@pytest.mark.parametrize("n_vectors", [2, 3])
@pytest.mark.parametrize("hidden_channels", [16])
@pytest.mark.parametrize("num_layers_mlp", [1])
@pytest.mark.parametrize("num_blocks,hidden_vectors", [(1, None), (2, 1), (2, 2)])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
@pytest.mark.parametrize("num_scalars,include_edges", [(0, True), (1, False)])
@pytest.mark.parametrize(
    "operation, fm_norm",
    [("diff", True), ("diff", False), ("add", True), ("add", False), ("single", False)],
)
@pytest.mark.parametrize("nonlinearity", ["softplus", "exp", "softmax", None])
def test_equivariance(
    batch_dims,
    jet_size,
    n_vectors,
    hidden_channels,
    num_layers_mlp,
    num_blocks,
    hidden_vectors,
    logm2_std,
    logm2_mean,
    include_edges,
    num_scalars,
    operation,
    nonlinearity,
    fm_norm,
):
    assert len(batch_dims) == 1
    dtype = torch.float64

    # construct sparse tensors containing a set of equal-multiplicity jets
    ptr = torch.arange(0, (batch_dims[0] + 1) * jet_size, jet_size)

    # input to mlp: only edge attributes
    calc_node_attr = lambda fm: torch.zeros(*fm.shape[:-1], num_scalars, dtype=dtype)
    equivectors = EquiMLP(
        n_vectors=n_vectors,
        num_scalars=num_scalars,
        hidden_channels=hidden_channels,
        num_layers_mlp=num_layers_mlp,
        num_blocks=num_blocks,
        hidden_vectors=hidden_vectors,
        include_edges=include_edges,
        operation=operation,
        nonlinearity=nonlinearity,
        fm_norm=fm_norm,
    ).to(dtype=dtype)

    fm = sample_particle(
        batch_dims + [jet_size], logm2_std, logm2_mean, dtype=dtype
    ).flatten(0, 1)

    # careful: same global transformation for each jet
    random = rand_lorentz(batch_dims, dtype=dtype)
    random = random.unsqueeze(1).repeat(1, jet_size, 1, 1).view(*fm.shape, 4)

    # path 1: global transform + predict vectors
    fm_prime = torch.einsum("...ij,...j->...i", random, fm)
    node_attr_prime = calc_node_attr(fm_prime)
    vecs_prime1 = equivectors(fourmomenta=fm_prime, scalars=node_attr_prime, ptr=ptr)

    # path 2: predict vectors + global transform
    node_attr = calc_node_attr(fm)
    vecs = equivectors(fourmomenta=fm, scalars=node_attr, ptr=ptr)
    vecs_prime2 = torch.einsum("...ij,...kj->...ki", random, vecs)

    # test that vectors are predicted equivariantly
    torch.testing.assert_close(vecs_prime1, vecs_prime2, **TOLERANCES)
